from qmctorch.wavefunction import Molecule
import numpy as np
import unittest


class TestMolecule(unittest.TestCase):

    def test1_create(self):

        # molecule
        mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        mol.print_total_energy()

    def test2_load(self):
        mol = Molecule(load='H2_pyscf_sto-3g.hdf5')

    def test3_domain(self):
        mol = Molecule(load='H2_pyscf_sto-3g.hdf5')

        domain_center = mol.domain('center')
        assert (domain_center['center'] ==
                np.array([0., 0., 0.5])).all()

        domain_uniform = mol.domain('uniform')
        assert domain_uniform == {'min': -0.5, 'max': 1.5}

        domain_normal = mol.domain('normal')
        assert np.all(domain_normal['mean']
                      == np.array([0., 0., 0.5]))

        domain_atomic = mol.domain('atomic')
        assert np.all(domain_atomic['atom_coords'] == np.array([[0., 0., 0.],
                                                                [0., 0., 1.]]))


if __name__ == "__main__":
    unittest.main()



import torch
from tqdm import tqdm
from torch.autograd import Variable
from .sampler_base import SamplerBase


class Hamiltonian(SamplerBase):

    def __init__(self, nwalkers=100, nstep=100, nelec=1, ndim=3,
                 step_size=0.1, ntherm=-1, ndecor=1, init={'min': -2, 'max': 2}, L=10,
                 with_tqdm=True):
        """Hamiltonian Monte Carlo Sampler.

        Args:
            nwalkers (int, optional): Number of walkers. Defaults to 100.
            nstep (int, optional): Number of steps. Defaults to 1000.
            step_size (int, optional): length of the step. Defaults to 3.
            nelec (int, optional): total number of electrons. Defaults to 1.
            ntherm (int, optional): number of mc step to thermalize. Defaults to -1, i.e. keep ponly last position
            ndecor (int, optional): number of mc step for decorelation. Defauts to 1.
            ndim (int, optional): total number of dimension. Defaults to 1.
            init (dict, optional): method to init the positions of the walkers. See Molecule.domain()
            L (int, optional): length of the trajectory . Defaults to 10.
            with_tqdm (bool, optional): use tqdm progress bar. Defaults to True.
        """

        SamplerBase.__init__(self, nwalkers, nstep,
                             step_size, ntherm, ndecor,
                             nelec, ndim, init, with_tqdm)
        self.traj_length = L

    @staticmethod
    def get_grad(func, inp):
        inp.requires_grad = True
        val = func(inp)
        z = Variable(torch.ones(val.shape))
        val.backward(z)
        fgrad = inp.grad.data
        inp.grad.data.zero_()
        inp.requires_grad = False
        return fgrad

    @staticmethod
    def log_func(func):
        return lambda x: -torch.log(func(x))

    def __call__(self, pdf, pos=None):
        """Generate walkers followinf HMC

        Arguments:
            pdf {callable} -- density to sample

        Keyword Arguments:
            ntherm {int} -- number of iterations needed to thermalize (default: {10})
            ndecor {int} -- number of iterations needed to decorelate (default: {10})
            pos {torch.tensor} -- initial position of the walker (default: {None})

        Returns:
            torch.tensor -- sampling points
        """

        if self.ntherm < 0:
            self.ntherm = self.nstep + self.ntherm

        self.walkers.initialize(pos=pos)
        self.walkers.pos = self.walkers.pos.clone()

        # get the logpdf function
        logpdf = self.log_func(pdf)

        pos = []
        rate = 0
        idecor = 0

        if self.with_tqdm:
            rng = tqdm(range(self.nstep))
        else:
            rng = range(self.nstep)

        for istep in rng:

            # move the walkers
            self.walkers.pos, _r = self._step(
                logpdf, self.get_grad, self.step_size, self.traj_length,
                self.walkers.pos)
            rate += _r

            # store
            if (istep >= self.ntherm):
                if (idecor % self.ndecor == 0):
                    pos.append(self.walkers.pos.detach())
                idecor += 1

        # print stats
        print("Acceptance rate %1.3f %%" % (rate / self.nstep * 100))
        return torch.cat(pos).requires_grad_()

    @staticmethod
    def _step(U, get_grad, epsilon, L, qinit):
        '''Propagates all the walkers over on traj
        Args:
            pdf (callable): the target dist
            get_grad (callable) : get the value of the target dist gradient
            epsilon (float) : step size
            L (int) : number of steps in the traj
            qinit (array) : initial positon of the walkers
        Returns:
            q : new positions of the walkers
            rate : accept rate
        '''

        # init the momentum
        q = qinit.clone()
        p = torch.randn(q.shape)

        # initial energy terms
        Einit = U(q) + 0.5 * (p**2).sum(1)

        # half step in momentum space
        p -= 0.5 * epsilon * get_grad(U, q)

        # full steps in q and p space
        for iL in range(L - 1):
            q = q + epsilon * p
            p -= 0.5 * epsilon * get_grad(U, q)

        # last full step in pos space
        q = q + epsilon * p

        # half step in momentum space
        p -= 0.5 * epsilon * get_grad(U, q)

        # negate momentum
        p = -p

        # current energy term
        Enew = U(q) + 0.5 * (p**2).sum(1)

        # metropolix accept/reject
        eps = torch.rand(Enew.shape)
        cond = (torch.exp(Einit - Enew) < eps).view(-1)
        q[cond] = qinit[cond]

        # comute the accept rate
        rate = cond.sum().float() / cond.shape[0]

        return q, rate



from qmctorch.wavefunction import Molecule
from qmctorch.wavefunction.orbital_configurations import OrbitalConfigurations
from qmctorch.utils import set_torch_double_precision
import numpy as np
import torch
import unittest
import itertools


class TestOrbitalConfiguration(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        self.orb_conf = OrbitalConfigurations(mol)

    def test_confs(self):

        self.orb_conf.get_configs('ground_state')
        self.orb_conf.get_configs('single(2,2)')
        self.orb_conf.get_configs('single_double(2,2)')
        self.orb_conf.get_configs('cas(2,2)')


if __name__ == "__main__":
    unittest.main()



import torch


def radial_slater(R, bas_n, bas_exp, xyz=None, derivative=0, jacobian=True):
    """Compute the radial part of STOs (or its derivative).

    Args:
        R (torch.tensor): distance between each electron and each atom
        bas_n (torch.tensor): principal quantum number
        bas_exp (torch.tensor): exponents of the exponential

    Keyword Arguments:
        xyz (torch.tensor): positions of the electrons
                            (needed for derivative) (default: {None})
        derivative (int): degree of the derivative (default: {0})
        jacobian (bool): return the jacobian, i.e the sum of the gradients
                           (default: {True})

    Returns:
        torch.tensor: values of each orbital radial part at each position
    """

    if derivative == 0:
        return R**bas_n * torch.exp(-bas_exp * R)

    elif derivative > 0:

        rn = R**(bas_n)
        nabla_rn = (bas_n * R**(bas_n - 2)).unsqueeze(-1) * xyz

        er = torch.exp(-bas_exp * R)
        nabla_er = -(bas_exp * er).unsqueeze(-1) * \
            xyz / R.unsqueeze(-1)

        if derivative == 1:

            if jacobian:
                nabla_rn = nabla_rn.sum(3)
                nabla_er = nabla_er.sum(3)
                return nabla_rn * er + rn * nabla_er
            else:
                return nabla_rn * \
                    er.unsqueeze(-1) + rn.unsqueeze(-1) * nabla_er

        elif derivative == 2:

            sum_xyz2 = (xyz**2).sum(3)

            lap_rn = bas_n * (3 * R**(bas_n - 2) +
                              sum_xyz2 * (bas_n - 2) * R**(bas_n - 4))

            lap_er = bas_exp**2 * er * sum_xyz2 / R**2 \
                - 2 * bas_exp * er * sum_xyz2 / R**3

            return lap_rn * er + 2 * \
                (nabla_rn * nabla_er).sum(3) + rn * lap_er


def radial_gaussian(R, bas_n, bas_exp, xyz=None, derivative=0, jacobian=True):
    """ompute the radial part of GTOs (or its derivative).

    Args:
        R (torch.tensor): distance between each electron and each atom
        bas_n (torch.tensor): principal quantum number
        bas_exp (torch.tensor): exponents of the exponential

    Keyword Arguments:
        xyz (torch.tensor): positions of the electrons
                            (needed for derivative) (default: {None})
        derivative (int): degree of the derivative (default: {0})
        jacobian (bool): return the jacobian, i.e the sum of the gradients
                           (default: {True})

    Returns:
        torch.tensor: values of each orbital radial part at each position
    """
    if derivative == 0:
        return R**bas_n * torch.exp(-bas_exp * R**2)

    elif derivative > 0:

        rn = R**(bas_n)
        nabla_rn = (bas_n * R**(bas_n - 2)).unsqueeze(-1) * xyz

        er = torch.exp(-bas_exp * R**2)
        nabla_er = -2 * (bas_exp * er).unsqueeze(-1) * xyz

        if derivative == 1:
            if jacobian:
                nabla_rn = nabla_rn.sum(3)
                nabla_er = nabla_er.sum(3)
                return nabla_rn * er + rn * nabla_er
            else:
                return nabla_rn * \
                    er.unsqueeze(-1) + rn.unsqueeze(-1) * nabla_er

        elif derivative == 2:

            lap_rn = bas_n * (3 * R**(bas_n - 2)
                              + (xyz**2).sum(3) * (bas_n - 2) * R**(bas_n - 4))

            lap_er = 4 * bas_exp**2 * (xyz**2).sum(3) * er \
                - 6 * bas_exp * er

            return lap_rn * er + 2 * \
                (nabla_rn * nabla_er).sum(3) + rn * lap_er



__all__ = ['SolverBase', 'SolverOrbital', 'SolverOrbitalHorovod']

from .solver_base import SolverBase
from .solver_orbital import SolverOrbital
from .solver_orbital_horovod import SolverOrbitalHorovod



import torch
from torch import nn

from .atomic_orbitals import AtomicOrbitals
from .slater_pooling import SlaterPooling
from .kinetic_pooling import KineticPooling
from .orbital_configurations import OrbitalConfigurations
from .wf_base import WaveFunction
from .jastrow import TwoBodyJastrowFactor

from ..utils import register_extra_attributes


class Orbital(WaveFunction):

    def __init__(self, mol, configs='ground_state',
                 kinetic='jacobi', use_jastrow=True, cuda=False):
        """Implementation of the QMC Network. 

        Args:
            mol (qmc.wavefunction.Molecule): a molecule object
            configs (str, optional): defines the CI configurations to be used. Defaults to 'ground_state'.
            kinetic (str, optional): method to compute the kinetic energy. Defaults to 'jacobi'.
            use_jastrow (bool, optional): turn jastrow factor ON/OFF. Defaults to True.
            cuda (bool, optional): turns GPU ON/OFF  Defaults to False.

        Examples::
            >>> mol = Molecule('h2o.xyz', calculator='adf', basis = 'dzp')
            >>> wf = Orbital(mol, configs='cas(2,2)')
        """

        super(Orbital, self).__init__(mol.nelec, 3, kinetic, cuda)

        # check for cuda
        if not torch.cuda.is_available and self.wf.cuda:
            raise ValueError('Cuda not available, use cuda=False')

        # number of atoms
        self.mol = mol
        self.atoms = mol.atoms
        self.natom = mol.natom

        # define the atomic orbital layer
        self.ao = AtomicOrbitals(mol, cuda)

        # define the mo layer
        self.mo_scf = nn.Linear(
            mol.basis.nao, mol.basis.nmo, bias=False)
        self.mo_scf.weight = self.get_mo_coeffs()
        self.mo_scf.weight.requires_grad = False
        if self.cuda:
            self.mo_scf.to(self.device)

        # define the mo mixing layer
        self.mo = nn.Linear(mol.basis.nmo, mol.basis.nmo, bias=False)
        self.mo.weight = nn.Parameter(torch.eye(mol.basis.nmo))
        if self.cuda:
            self.mo.to(self.device)

        # jastrow
        self.use_jastrow = use_jastrow
        self.jastrow = TwoBodyJastrowFactor(mol.nup, mol.ndown,
                                            w=1., cuda=cuda)

        # define the SD we want
        self.orb_confs = OrbitalConfigurations(mol)
        self.configs_method = configs
        self.configs = self.orb_confs.get_configs(configs)
        self.nci = len(self.configs[0])

        #  define the SD pooling layer
        self.pool = SlaterPooling(
            self.configs, mol, cuda)

        # pooling operation to directly compute
        # the kinetic energies via Jacobi formula
        self.kinpool = KineticPooling(
            self.configs, mol, cuda)

        # define the linear layer
        self.fc = nn.Linear(self.nci, 1, bias=False)
        self.fc.weight.data.fill_(0.)
        self.fc.weight.data[0][0] = 1.

        if self.cuda:
            self.fc = self.fc.to(self.device)

        if kinetic == 'jacobi':
            self.local_energy = self.local_energy_jacobi

        if self.cuda:
            self.device = torch.device('cuda')
            self.to(self.device)

        # register the callable for hdf5 dump
        register_extra_attributes(self,
                                  ['ao', 'mo_scf',
                                   'mo', 'jastrow',
                                   'pool', 'kinpool', 'fc'])

    def get_mo_coeffs(self):
        mo_coeff = torch.tensor(self.mol.basis.mos).type(
            torch.get_default_dtype())
        return nn.Parameter(mo_coeff.transpose(0, 1).contiguous())

    def update_mo_coeffs(self):
        self.mol.atom_coords = self.ao.atom_coords.detach().numpy().tolist()
        self.mo.weight = self.get_mo_coeffs()

    def forward(self, x, ao=None):
        """computes the value of the wave function for the sampling points

        .. math::
            \Psi(R) = \sum_{n} c_n D^{u}_n(r^u) \\times D^{d}_n(r^d)

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            ao (torch.tensor, optional): values of the atomic orbitals (Nbatch, Nelec, Nao)

        Returns:
            torch.tensor: values of the wave functions at each sampling point (Nbatch, 1)

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis = 'dzp')
            >>> wf = Orbital(mol, configs='cas(2,2)')
            >>> pos = torch.rand(500,6)
            >>> vals = wf(pos)
        """

        if self.use_jastrow:
            J = self.jastrow(x)

        # atomic orbital
        if ao is None:
            x = self.ao(x)

        else:
            x = ao

        # molecular orbitals
        x = self.mo_scf(x)

        # mix the mos
        x = self.mo(x)

        # pool the mos
        x = self.pool(x)

        if self.use_jastrow:
            return J * self.fc(x)

        else:
            return self.fc(x)

    def _get_mo_vals(self, x, derivative=0):
        """Get the values of MOs

        Arguments:
            x {torch.tensor} -- positions of the electrons [nbatch, nelec*ndim]

        Keyword Arguments:
            derivative {int} -- order of the derivative (default: {0})

        Returns:
            torch.tensor -- MO matrix [nbatch, nelec, nmo]
        """
        return self.mo(self.mo_scf(self.ao(x, derivative=derivative)))

    def local_energy_jacobi(self, pos):
        """Computes the local energy using the Jacobi formula

        .. math::
            E = K(R) + V_{ee}(R) + V_{en}(R) + V_{nn}

        Args:
            pos (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            [torch.tensor]: values of the local enrgies at each sampling points

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis = 'dzp')
            >>> wf = Orbital(mol, configs='cas(2,2)')
            >>> pos = torch.rand(500,6)
            >>> vals = wf.local_energy_jacobi(pos)

        """

        ke = self.kinetic_energy_jacobi(pos)

        return ke \
            + self.nuclear_potential(pos) \
            + self.electronic_potential(pos) \
            + self.nuclear_repulsion()

    def kinetic_energy_jacobi(self, x, **kwargs):
        """Compute the value of the kinetic enery using the Jacobi Formula.
        C. Filippi, Simple Formalism for Efficient Derivatives .

        .. math::
             \\frac{K(R)}{\Psi(R)} = Tr(A^{-1} B_{kin})

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the kinetic energy at each sampling points
        """

        mo = self._get_mo_vals(x)
        d2mo = self._get_mo_vals(x, derivative=2)
        djast_dmo, d2jast_mo = None, None

        if self.use_jastrow:

            jast = self.jastrow(x)
            djast = self.jastrow(x, derivative=1, jacobian=False)
            djast = djast.transpose(1, 2) / jast.unsqueeze(-1)

            dao = self.ao(x, derivative=1,
                          jacobian=False).transpose(2, 3)
            dmo = self.mo(self.mo_scf(dao)).transpose(2, 3)
            djast_dmo = (djast.unsqueeze(2) * dmo).sum(-1)

            d2jast = self.jastrow(x, derivative=2) / jast
            d2jast_mo = d2jast.unsqueeze(-1) * mo

        kin, psi = self.kinpool(mo, d2mo, djast_dmo, d2jast_mo)

        return self.fc(kin) / self.fc(psi)

    def nuclear_potential(self, pos):
        """Computes the electron-nuclear term

        .. math:
            V_{en} = - \sum_e \sum_n \\frac{Z_n}{r_{en}}

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the electon-nuclear energy at each sampling points
        """

        p = torch.zeros(pos.shape[0], device=self.device)
        for ielec in range(self.nelec):
            istart = ielec * self.ndim
            iend = (ielec + 1) * self.ndim
            pelec = pos[:, istart:iend]
            for iatom in range(self.natom):
                patom = self.ao.atom_coords[iatom, :]
                Z = self.ao.atomic_number[iatom]
                r = torch.sqrt(((pelec - patom)**2).sum(1))  # + 1E-12
                p += -Z / r
        return p.view(-1, 1)

    def electronic_potential(self, pos):
        """Computes the electron-electron term

        .. math:
            V_{ee} = \sum_{e_1} \sum_{e_2} \\frac{1}{r_{e_1e_2}}

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the electon-electron energy at each sampling points
        """

        pot = torch.zeros(pos.shape[0], device=self.device)

        for ielec1 in range(self.nelec - 1):
            epos1 = pos[:, ielec1 *
                        self.ndim:(ielec1 + 1) * self.ndim]
            for ielec2 in range(ielec1 + 1, self.nelec):
                epos2 = pos[:, ielec2 *
                            self.ndim:(ielec2 + 1) * self.ndim]
                r = torch.sqrt(((epos1 - epos2)**2).sum(1))  # + 1E-12
                pot += (1. / r)
        return pot.view(-1, 1)

    def nuclear_repulsion(self):
        """Computes the nuclear-nuclear repulsion term

        .. math:
            V_{nn} = \sum_{n_1} \sum_{n_2} \\frac{Z_1Z_2}{r_{n_1n_2}}

        Returns:
            torch.tensor: values of the nuclear-nuclear energy at each sampling points
        """

        vnn = 0.
        for at1 in range(self.natom - 1):
            c0 = self.ao.atom_coords[at1, :]
            Z0 = self.ao.atomic_number[at1]
            for at2 in range(at1 + 1, self.natom):
                c1 = self.ao.atom_coords[at2, :]
                Z1 = self.ao.atomic_number[at2]
                rnn = torch.sqrt(((c0 - c1)**2).sum())
                vnn += Z0 * Z1 / rnn
        return vnn

    def geometry(self, pos):
        """Returns the gemoetry of the system in xyz format

        Args:
            pos (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            list: list where each element is one line of the xyz file
        """
        d = []
        for iat in range(self.natom):
            at = self.atoms[iat]
            xyz = self.ao.atom_coords[iat,
                                      :].detach().numpy().tolist()
            d.append((at, xyz))
        return d



from qmctorch.wavefunction import Molecule, Orbital
from qmctorch.sampler import Metropolis, Hamiltonian, GeneralizedMetropolis
from qmctorch.sampler.walkers import Walkers
from qmctorch.utils import set_torch_double_precision
import numpy as np
import torch
import unittest
import itertools


class TestSampler(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        self.mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # orbital
        self.wf = Orbital(self.mol)

    def test_walkers_init(self):
        """Test different initialization methods of the walkers."""
        w1 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('center'))

        w2 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('uniform'))

        w3 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('normal'))

        w4 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('atomic'))

    def test_metropolis(self):
        """Test Metropolis sampling."""

        sampler = Metropolis(
            nwalkers=10,
            nstep=20,
            step_size=0.5,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'))

        for m in ['one-elec', 'all-elec', 'all-elec-iter']:
            for p in ['normal', 'uniform']:

                sampler.configure_move({'type': m, 'proba': p})
                pos = sampler(self.wf.pdf)

    def test_hmc(self):
        """Test HMC sampler."""
        sampler = Hamiltonian(
            nwalkers=10,
            nstep=20,
            step_size=0.1,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'))

        pos = sampler(self.wf.pdf)

    def test_gmh(self):
        """Test generalized MH."""
        sampler = GeneralizedMetropolis(
            nwalkers=10, nstep=20, step_size=0.2,
            nelec=self.wf.nelec, ndim=self.wf.ndim,
            init=self.mol.domain('normal'))

        pos = sampler(self.wf.pdf)


if __name__ == "__main__":
    unittest.main()



import torch
from torch.utils.data import DataLoader
import warnings

from .solver_base import SolverBase
from qmctorch.utils import (DataSet, Loss, OrthoReg)

try:
    import horovod.torch as hvd
except ModuleNotFoundError:
    pass


def printd(rank, *args):
    if rank == 0:
        print(*args)


class SolverOrbitalHorovod(SolverBase):

    def __init__(self, wf=None, sampler=None, optimizer=None, scheduler=None):
        """Horovod distributed solver

        Keyword Arguments:
            wf {WaveFunction} -- WaveFuntion object (default: {None})
            sampler {SamplerBase} -- Samppler (default: {None})
            optimizer {torch.optim} -- Optimizer (default: {None})
            scheduler (torch.schedul) -- Scheduler (default: {None})
        """

        SolverBase.__init__(self, wf, sampler, optimizer)

        hvd.broadcast_optimizer_state(self.opt, root_rank=0)
        self.opt = hvd.DistributedOptimizer(
            self.opt, named_parameters=self.wf.named_parameters())

        self.sampler.nwalkers //= hvd.size()
        self.sampler.walkers.nwalkers //= hvd.size()

    def run(self, nepoch, batchsize=None, loss='variance', num_threads=1):
        '''Train the model.

        Arg:
            nepoch : number of epoch
            batchsize : size of the minibatch, if None take all points at once
            loss : loss used ('energy','variance' or callable (for supervised)
        '''

        if 'lpos_needed' not in self.opt.defaults:
            self.opt.defaults['lpos_needed'] = False

        self.wf.train()

        hvd.broadcast_parameters(self.wf.state_dict(), root_rank=0)
        torch.set_num_threads(num_threads)

        # sample the wave function
        pos = self.sampler(ntherm=self.resample.ntherm)
        pos.requires_grad_(False)

        # handle the batch size
        if batchsize is None:
            batchsize = len(pos)

        # change the number of steps
        _nstep_save = self.sampler.nstep
        self.sampler.nstep = self.resample.resample

        # create the data loader
        self.dataset = DataSet(pos)

        if self.cuda:
            kwargs = {'num_workers': num_threads, 'pin_memory': True}
        else:
            kwargs = {'num_workers': num_threads}

        self.dataloader = DataLoader(self.dataset,
                                     batch_size=batchsize,
                                     **kwargs)

        # get the loss
        self.loss = Loss(self.wf, method=loss)

        # orthogonalization penalty for the MO coeffs
        self.ortho_loss = OrthoReg()

        cumulative_loss = []
        min_loss = 1E3

        # get the initial observalbe

        self.get_observable(self.obs_dict, pos)
        for n in range(nepoch):

            printd(
                hvd.rank(),
                '----------------------------------------')
            printd(hvd.rank(), 'epoch %d' % n)

            cumulative_loss = 0.
            for data in self.dataloader:

                lpos = data.to(self.device)
                lpos.requires_grad = True

                loss, eloc = self.loss(lpos)
                if self.wf.mo.weight.requires_grad:
                    loss += self.ortho_loss(self.wf.mo.weight)
                cumulative_loss += loss

                # compute gradients
                self.opt.zero_grad()
                loss.backward()

                # optimize
                if 'lpos_needed' in self.opt.defaults:
                    self.opt.step(lpos)
                else:
                    self.opt.step()

            cumulative_loss = self.metric_average(
                cumulative_loss, 'cum_loss')
            if hvd.rank() == 0:
                if cumulative_loss < min_loss:
                    min_loss = self.save_checkpoint(
                        n, cumulative_loss, self.save_model)

            self.get_observable(self.obs_dict, pos, local_energy=eloc)
            self.print_observable(cumulative_loss)

            printd(
                hvd.rank(),
                '----------------------------------------')

            # resample the data
            if (n % self.resample.resample_every == 0) or (
                    n == nepoch - 1):
                if self.resample.resample_from_last:
                    pos = pos.clone().detach().to(self.device)
                else:
                    pos = None

                pos = self.sampler(pos=pos,
                                   ntherm=self.resample.ntherm,
                                   with_tqdm=False)
                pos.requires_grad_(False)

                self.dataloader.dataset.data = pos

            if self.task == 'geo_opt':
                self.wf.update_mo_coeffs()

            if self.scheduler is not None:
                self.scheduler.step()

        # restore the sampler number of step
        self.sampler.nstep = _nstep_save

    def single_point(self, pos=None, prt=True, ntherm=-1, ndecor=100):
        """Performs a single point calculation

        Keyword Arguments:
            pos {torch.tensor} -- positions of the walkers If none, sample
                                  (default: {None})
            prt {bool} -- print energy/variance values (default: {True})
            ntherm {int} -- number of MC steps to thermalize (default: {-1})
            ndecor {int} -- number of MC step to decorelate  (default: {100})

        Returns:
            tuple -- (position, energy, variance)
        """

        self.wf.eval()
        num_threads = 1
        hvd.broadcast_parameters(self.wf.state_dict(), root_rank=0)
        torch.set_num_threads(num_threads)

        # sample the wave function
        pos = self.sampler(ntherm=self.resample.ntherm)
        pos.requires_grad_(False)

        # handle the batch size
        batchsize = len(pos)

        # create the data loader
        self.dataset = DataSet(pos)

        if self.cuda:
            kwargs = {'num_workers': num_threads, 'pin_memory': True}
        else:
            kwargs = {'num_workers': num_threads}

        self.dataloader = DataLoader(self.dataset,
                                     batch_size=batchsize,
                                     **kwargs)

        for data in self.dataloader:

            lpos = data.to(self.device)
            lpos.requires_grad = True
            eloc = self.wf.local_energy(lpos)

        eloc_all = hvd.allgather(eloc, name='local_energies')
        e = torch.mean(eloc_all)
        s = torch.var(eloc_all)

        if prt:
            printd(hvd.rank(), 'Energy   : ', e)
            printd(hvd.rank(), 'Variance : ', s)
        return pos, e, s

    @staticmethod
    def metric_average(val, name):
        """Average a give quantity over all processes

        Arguments:
            val {torch.tensor} -- data to average
            name {str} -- name of the data

        Returns:
            torch.tensor -- Averaged quantity
        """
        tensor = val.clone().detach()
        avg_tensor = hvd.allreduce(tensor, name=name)
        return avg_tensor.item()



import torch
from .walkers import Walkers


class SamplerBase(object):

    def __init__(self, nwalkers, nstep, step_size, ntherm, ndecor, nelec, ndim, init, with_tqdm):
        """Base class for the sampler.

        Arguments:
            nwalkers {int]} -- number of walkers
            nstep {[int]} -- number of MC step
            step_size {[float]} -- size of the MC step
            nelec {int} -- number of electrons
            ndim {int} -- number of dimension per elec
            init {dict} -- method/data to initialize th walkers
        """

        self.nwalkers = nwalkers
        self.nelec = nelec
        self.ndim = ndim
        self.nstep = nstep
        self.step_size = step_size
        self.ntherm = ntherm
        self.ndecor = ndecor
        self.cuda = False
        self.device = torch.device('cpu')
        self.with_tqdm = with_tqdm

        self.walkers = Walkers(
            nwalkers=nwalkers, nelec=nelec, ndim=ndim, init=init)

    def __call__(self, pdf, *args, **kwargs):
        raise NotImplementedError(
            "Sampler must have a __call__ method")



import torch
from torch.utils.data import DataLoader

from .solver_base import SolverBase
from qmctorch.utils import (
    DataSet, Loss, OrthoReg, dump_to_hdf5, add_group_attr)


class SolverOrbital(SolverBase):

    def __init__(self, wf=None, sampler=None, optimizer=None,
                 scheduler=None, output=None):
        """Basic QMC solver 

        Args:
            wf (qmctorch.WaveFunction, optional): wave function. Defaults to None.
            sampler (qmctorch.sampler, optional): Sampler. Defaults to None.
            optimizer (torch.optim, optional): optimizer. Defaults to None.
            scheduler (torch.optim, optional): scheduler. Defaults to None.
            output (str, optional): hdf5 filename. Defaults to None.
        """

        SolverBase.__init__(self, wf, sampler, optimizer, output)

    def run(self, nepoch, batchsize=None, loss='energy',
            clip_loss=False, grad='manual', hdf5_group=None):
        """Run th optimization

        Args:
            nepoch (int): Number of optimziation step
            batchsize (int, optional): Number of sample in a mini batch.
                                       If None, all samples are used.
                                       Defaults to None.
            loss (str, optional): merhod to compute the loss: variance or energy.
                                  Defaults to 'energy'.
            clip_loss (bool, optional): Clip the loss values at +/- 5std.
                                        Defaults to False.
            grad (str, optional): method to compute the gradients: 'auto' or 'manual'.
                                  Defaults to 'auto'.
            hdf5_group (str, optional): name of the hdf5 group where to store the data.
                                        Defaults to wf.task.
        """

        # observalbe
        if not hasattr(self, 'observable'):
            self.track_observable(['local_energy'])

        self.evaluate_gradient = {
            'auto': self.evaluate_grad_auto,
            'manual': self.evaluate_grad_manual}[grad]

        if 'lpos_needed' not in self.opt.__dict__.keys():
            self.opt.lpos_needed = False

        # sample the wave function
        pos = self.sampler(self.wf.pdf)

        # get the loss
        self.loss = Loss(self.wf, method=loss, clip=clip_loss)
        self.loss.use_weight = (
            self.resampling_options.resample_every > 1)

        # orthogonalization penalty for the MO coeffs
        self.ortho_loss = OrthoReg()

        # handle the batch size
        if batchsize is None:
            batchsize = len(pos)

        # change the number of steps/walker size
        _nstep_save = self.sampler.nstep
        _ntherm_save = self.sampler.ntherm
        _nwalker_save = self.sampler.walkers.nwalkers
        if self.resampling_options.mode == 'update':
            self.sampler.ntherm = -1
            self.sampler.nstep = self.resampling_options.nstep_update
            self.sampler.walkers.nwalkers = pos.shape[0]
            self.sampler.nwalkers = pos.shape[0]

        # create the data loader
        self.dataset = DataSet(pos)
        self.dataloader = DataLoader(
            self.dataset, batch_size=batchsize)

        cumulative_loss = []
        min_loss = 1E3

        # get the initial observalbe
        self.store_observable(pos)

        # loop over the epoch
        for n in range(nepoch):
            print('----------------------------------------')
            print('epoch %d' % n)

            cumulative_loss = 0

            # loop over the batches
            for ibatch, data in enumerate(self.dataloader):

                # port data to device
                lpos = data.to(self.device)

                # get the gradient
                loss, eloc = self.evaluate_gradient(lpos)
                cumulative_loss += loss

                # optimize the parameters
                self.optimization_step(lpos)

                # observable
                self.store_observable(
                    pos, local_energy=eloc, ibatch=ibatch)

            # save the model if necessary
            if cumulative_loss < min_loss:
                min_loss = self.save_checkpoint(
                    n, cumulative_loss, self.save_model)

            self.print_observable(cumulative_loss)

            print('----------------------------------------')

            # resample the data
            pos = self.resample(n, pos)

            if self.task == 'geo_opt':
                self.wf.update_mo_coeffs()

            if self.scheduler is not None:
                self.scheduler.step()

        # restore the sampler number of step
        self.sampler.nstep = _nstep_save
        self.sampler.ntherm = _ntherm_save
        self.sampler.walkers.nwalkers = _nwalker_save
        self.sampler.nwalkers = _nwalker_save

        # dump
        if hdf5_group is None:
            hdf5_group = self.task
        dump_to_hdf5(self.observable, self.hdf5file, hdf5_group)
        add_group_attr(self.hdf5file, hdf5_group, {'type': 'opt'})

        return self.observable

    def evaluate_grad_auto(self, lpos):
        """Evaluate the gradient using automatic differentiation

        Args:
            lpos (torch.tensor): sampling points

        Returns:
            tuple: loss values and local energies
        """

        # compute the loss
        loss, eloc = self.loss(lpos)

        # add mo orthogonalization if required
        if self.wf.mo.weight.requires_grad and self.ortho_mo:
            loss += self.ortho_loss(self.wf.mo.weight)

        # compute local gradients
        self.opt.zero_grad()
        loss.backward()

        return loss, eloc

    def evaluate_grad_manual(self, lpos):
        """Evaluate the gradient using low variance express

        Args:
            lpos ([type]): [description]

        Args:
            lpos (torch.tensor): sampling points

        Returns:
            tuple: loss values and local energies
        """

        if self.loss.method in ['energy', 'weighted-energy']:

            ''' Get the gradient of the total energy
            dE/dk = < (dpsi/dk)/psi (E_L - <E_L >) >
            '''

            # compute local energy and wf values
            _, eloc = self.loss(lpos, no_grad=True)
            psi = self.wf(lpos)
            norm = 1. / len(psi)

            # evaluate the prefactor of the grads
            weight = eloc.clone()
            weight -= torch.mean(eloc)
            weight /= psi
            weight *= 2.
            weight *= norm

            # compute the gradients
            self.opt.zero_grad()
            psi.backward(weight)

            return torch.mean(eloc), eloc

        else:
            raise ValueError(
                'Manual gradient only for energy minimization')



import torch
from torch import nn
from .radial_functions import radial_gaussian, radial_slater
from .norm_orbital import atomic_orbital_norm
from .spherical_harmonics import Harmonics
from ..utils import register_extra_attributes


class AtomicOrbitals(nn.Module):

    def __init__(self, mol, cuda=False):
        """Computes the value of atomic orbitals

        Args:
            mol (Molecule): Molecule object
            cuda (bool, optional): Turn GPU ON/OFF Defaults to False.
        """

        super(AtomicOrbitals, self).__init__()
        dtype = torch.get_default_dtype()

        # wavefunction data
        self.nelec = mol.nelec
        self.norb = mol.basis.nao
        self.ndim = 3

        # make the atomic position optmizable
        self.atom_coords = nn.Parameter(torch.tensor(
            mol.basis.atom_coords_internal).type(dtype))
        self.atom_coords.requires_grad = True
        self.natoms = len(self.atom_coords)
        self.atomic_number = mol.atomic_number

        # define the BAS positions.
        self.nshells = torch.tensor(mol.basis.nshells)
        self.bas_coords = self.atom_coords.repeat_interleave(
            self.nshells, dim=0)
        self.nbas = len(self.bas_coords)

        # index for the contractions
        self.index_ctr = torch.tensor(mol.basis.index_ctr)

        # get the coeffs of the bas
        self.bas_coeffs = torch.tensor(
            mol.basis.bas_coeffs).type(dtype)

        # get the exponents of the bas
        self.bas_exp = nn.Parameter(
            torch.tensor(mol.basis.bas_exp).type(dtype))
        self.bas_exp.requires_grad = True

        # harmonics generator
        if mol.basis.harmonics_type == 'sph':
            self.bas_n = torch.tensor(mol.basis.bas_n).type(dtype)
            self.harmonics = Harmonics(
                mol.basis.harmonics_type,
                bas_l=mol.basis.bas_l,
                bas_m=mol.basis.bas_m)

        elif mol.basis.harmonics_type == 'cart':
            self.bas_n = torch.tensor(mol.basis.bas_kr).type(dtype)
            self.harmonics = Harmonics(
                mol.basis.harmonics_type,
                bas_kx=mol.basis.bas_kx,
                bas_ky=mol.basis.bas_ky,
                bas_kz=mol.basis.bas_kz)

        # select the radial apart
        radial_dict = {'sto': radial_slater,
                       'gto': radial_gaussian}
        self.radial = radial_dict[mol.basis.radial_type]

        # get the normalisation constants
        if hasattr(mol.basis, 'bas_norm') and False:
            self.norm_cst = torch.tensor(
                mol.basis.bas_norm).type(dtype)
        else:
            with torch.no_grad():
                self.norm_cst = atomic_orbital_norm(
                    mol.basis).type(dtype)

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self._to_device()

    def _to_device(self):
        """Export the non parameter variable to the device."""

        self.device = torch.device('cuda')
        self.to(self.device)
        attrs = ['bas_n', 'bas_coeffs',
                 'nshells', 'norm_cst', 'index_ctr']
        for at in attrs:
            self.__dict__[at] = self.__dict__[at].to(self.device)

    def forward(self, input, derivative=0, jacobian=True, one_elec=False):
        """Computes the values of the atomic orbitals.

        .. math::
            \phi_i(r_j) = \sum_n c_n \\text{Rad}^{i}_n(r_j) \\text{Y}^{i}_n(r_j)

        where Rad is the radial part and Y the spherical harmonics part. 
        It is also possible to compute the first and second derivatives

        .. math::
            \\nabla \phi_i(r_j) = \\frac{d}{dx_j} \phi_i(r_j) + \\frac{d}{dy_j} \phi_i(r_j) + \\frac{d}{dz_j} \phi_i(r_j) \n
            \\text{grad} \phi_i(r_j) = (\\frac{d}{dx_j} \phi_i(r_j), \\frac{d}{dy_j} \phi_i(r_j), \\frac{d}{dz_j} \phi_i(r_j)) \n
            \Delta \phi_i(r_j) = \\frac{d^2}{dx^2_j} \phi_i(r_j) + \\frac{d^2}{dy^2_j} \phi_i(r_j) + \\frac{d^2}{dz^2_j} \phi_i(r_j)

        Args:
            input (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                                        Defaults to 0.
            jacobian (bool, optional): Return the jacobian (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

            one_elec (bool, optional): if only one electron is in input

        Returns:
            torch.tensor: Value of the AO (or their derivatives) \n
                          size : Nbatch, Nelec, Norb (jacobian = True) \n
                          size : Nbatch, Nelec, Norb, Ndim (jacobian = False)

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> ao = AtomicOrbitals(mol)
            >>> pos = torch.rand(100,6)
            >>> aovals = ao(pos)
            >>> daovals = ao(pos,derivative=1)
        """

        if not jacobian:
            assert(derivative == 1)

        if one_elec:
            nelec_save = self.nelec
            self.nelec = 1

        nbatch = input.shape[0]

        # get the pos of the bas
        self.bas_coords = self.atom_coords.repeat_interleave(
            self.nshells, dim=0)

        # get the x,y,z, distance component of each point from each RBF center
        # -> (Nbatch,Nelec,Nbas,Ndim)
        xyz = (input.view(-1, self.nelec, 1, self.ndim) -
               self.bas_coords[None, ...])
        # print('xyz : ', time()-t0)

        # compute the distance
        # -> (Nbatch,Nelec,Nbas)
        r = torch.sqrt((xyz**2).sum(3))

        # radial part
        # -> (Nbatch,Nelec,Nbas)
        R = self.radial(r, self.bas_n, self.bas_exp)

        # compute by the spherical harmonics
        # -> (Nbatch,Nelec,Nbas)
        Y = self.harmonics(xyz)

        # values of AO
        # -> (Nbatch,Nelec,Nbas)
        if derivative == 0:
            bas = R * Y

        # values of first derivative
        elif derivative == 1:

            # return the jacobian
            if jacobian:
                dR = self.radial(
                    r, self.bas_n, self.bas_exp, xyz=xyz, derivative=1)
                dY = self.harmonics(xyz, derivative=1)

                # -> (Nbatch,Nelec,Nbas)
                bas = dR * Y + R * dY

            # returm individual components
            else:
                dR = self.radial(
                    r, self.bas_n, self.bas_exp, xyz=xyz, derivative=1, jacobian=False)
                dY = self.harmonics(xyz, derivative=1, jacobian=False)
                # -> (Nbatch,Nelec,Nbas,Ndim)
                bas = dR * Y.unsqueeze(-1) + R.unsqueeze(-1) * dY

        # second derivative
        elif derivative == 2:

            dR = self.radial(r, self.bas_n, self.bas_exp,
                             xyz=xyz, derivative=1, jacobian=False)
            dY = self.harmonics(xyz, derivative=1, jacobian=False)

            d2R = self.radial(
                r, self.bas_n, self.bas_exp, xyz=xyz, derivative=2)
            d2Y = self.harmonics(xyz, derivative=2)

            bas = d2R * Y + 2. * (dR * dY).sum(3) + R * d2Y

        # product with coefficients and primitives norm
        if jacobian:

            # -> (Nbatch,Nelec,Nbas)
            bas = self.norm_cst * self.bas_coeffs * bas

            # contract the basis
            # -> (Nbatch,Nelec,Norb)
            ao = torch.zeros(nbatch, self.nelec,
                             self.norb, device=self.device)
            ao.index_add_(2, self.index_ctr, bas)

        else:
            # -> (Nbatch,Nelec,Nbas, Ndim)
            bas = self.norm_cst.unsqueeze(-1) * \
                self.bas_coeffs.unsqueeze(-1) * bas

            # contract the basis
            # -> (Nbatch,Nelec,Norb, Ndim)
            ao = torch.zeros(nbatch, self.nelec, self.norb,
                             3, device=self.device)
            ao.index_add_(2, self.index_ctr, bas)

        if one_elec:
            self.nelec = nelec_save

        return ao

    def update(self, ao, pos, idelec):
        """Update an AO matrix with the new positions of one electron

        Args:
            ao (torch.tensor): initial AO matrix
            pos (torch.tensor): new positions of some electrons
            idelec (int): index of the electron that has moved

        Returns:
            torch.tensor: new AO matrix

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> ao = AtomicOrbitals(mol)
            >>> pos = torch.rand(100,6)
            >>> aovals = ao(pos)
            >>> id = 0
            >>> pos[:,:3] = torch.rand(100,3)
            >>> ao.update(aovals, pos, 0)
        """

        ao_new = ao.clone()
        ids, ide = (idelec) * 3, (idelec + 1) * 3
        ao_new[:, idelec, :] = self.forward(
            pos[:, ids:ide], one_elec=True).squeeze(1)
        return ao_new



import os
import math
import numpy as np
from mendeleev import element
from types import SimpleNamespace
import h5py

from .calculator.adf import CalculatorADF
from .calculator.pyscf import CalculatorPySCF

from ..utils import dump_to_hdf5, load_from_hdf5


class Molecule(object):

    def __init__(self, atom=None, calculator='adf',
                 scf='hf', basis='dzp', unit='bohr',
                 name=None, load=None):
        """Create a molecule in QMCTorch

        Args:
            atom (str or None, optional): defines the atoms and their positions. Defaults to None.
            calculator (str, optional): selet scf calculator. Defaults to 'adf'.
            scf (str, optional): select scf level of theory. Defaults to 'hf'.
            basis (str, optional): select the basis set. Defaults to 'dzp'.
            unit (str, optional): units of the coordinates. Defaults to 'bohr'.
            name (str or None, optional): name of the molecule. Defaults to None.
            load (str or None, optional): path to a hdf5 file to load. Defaults to None.

        Examples:
            >>> from qmctorch.wavefunction import Molecule
            >>> mol = Molecule(atom='H 0 0 0; H 0 0 1', unit='angs',
            ...                calculator='adf', basis='dzp')
        """

        self.atom_coords = []
        self.atomic_nelec = []
        self.atomic_number = []
        self.atoms = []
        self.atoms_str = atom
        self.hdf5file = None
        self.max_angular = 2
        self.name = name
        self.natom = 0
        self.ndown = 0
        self.nelec = 0
        self.nup = 0
        self.unit = unit
        self.basis = SimpleNamespace()

        if load is not None:
            print('Restarting calculation from ', load)
            self._load_hdf5(load)

        else:

            if self.unit not in ['angs', 'bohr']:
                raise ValueError('unit should be angs or bohr')

            self._process_atom_str()

            self.hdf5file = '_'.join(
                [self.name, calculator, basis]) + '.hdf5'

            if os.path.isfile(self.hdf5file):
                print('Reusing scf calculation from ', self.hdf5file)
                self.basis = self._load_basis()

            else:
                print('Running scf calculation')

                calc = {'adf': CalculatorADF,
                        'pyscf': CalculatorPySCF}[calculator]

                self.calculator = calc(
                    self.atoms, self.atom_coords, basis, scf, self.unit, self.name)

                self.basis = self.calculator.run()

                dump_to_hdf5(self, self.hdf5file,
                             root_name='molecule')

        self._check_basis()

    def domain(self, method):
        """Returns information to initialize the walkers

        Args:
            method (str): 'center', all electron at the center of the system
                          'uniform', all electrons in a cube surrounding the molecule
                          'normal',  all electrons in a sphere surrounding the molecule
                          'atomic', electrons around the atoms

        Returns:
            dict: dictionary containing corresponding information

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> domain = mol.domain('atomic')
        """
        domain = dict()

        if method == 'center':
            domain['center'] = np.mean(self.atom_coords, 0)

        elif method == 'uniform':
            domain['min'] = np.min(self.atom_coords) - 0.5
            domain['max'] = np.max(self.atom_coords) + 0.5

        elif method == 'normal':
            domain['mean'] = np.mean(self.atom_coords, 0)
            domain['sigma'] = np.diag(
                np.std(self.atom_coords, 0) + 0.25)

        elif method == 'atomic':
            domain['atom_coords'] = self.atom_coords
            domain['atom_num'] = self.atomic_number
            domain['atom_nelec'] = self.atomic_nelec

        else:
            raise ValueError(
                'Method to initialize the walkers not recognized')

        return domain

    def _process_atom_str(self):
        """Process the atom description."""

        if os.path.isfile(self.atoms_str):
            atoms = self._read_xyz_file()
        else:
            atoms = self.atoms_str.split(';')

        self._get_atomic_properties(atoms)

    def _get_atomic_properties(self, atoms):
        """Generates the atomic propeties of the molecule

        Args:
            atoms (str): atoms given in input
        """

        # loop over all atoms
        for a in atoms:
            atom_data = a.split()
            self.atoms.append(atom_data[0])
            x, y, z = float(atom_data[1]), float(
                atom_data[2]), float(atom_data[3])

            conv2bohr = 1
            if self.unit == 'angs':
                conv2bohr = 1.88973
            self.atom_coords.append(
                [x * conv2bohr, y * conv2bohr, z * conv2bohr])

            self.atomic_number.append(
                element(atom_data[0]).atomic_number)
            self.atomic_nelec.append(element(atom_data[0]).electrons)
            self.nelec += element(atom_data[0]).electrons

        # size of the system
        self.natom = len(self.atoms)
        if self.nelec % 2 != 0:
            raise ValueError("Only equal spin up/down supported.")
        self.nup = math.ceil(self.nelec / 2)
        self.ndown = math.floor(self.nelec / 2)

        # name of the system
        if self.name is None:
            self.name = self._get_mol_name(self.atoms)
        self.atoms = np.array(self.atoms)

    def _read_xyz_file(self):
        """Process a xyz file containing the data

        Returns:
            list -- atoms and xyz position
        """
        with open(self.atoms_str, 'r') as f:
            data = f.readlines()
        atoms = data[2:]
        self.atoms_str = ''
        for a in atoms[:-1]:
            self.atoms_str += a + '; '
        self.atoms_str += atoms[-1]
        return atoms

    @staticmethod
    def _get_mol_name(atoms):
        mol_name = ''
        unique_atoms = list(set(atoms))
        for ua in unique_atoms:
            mol_name += ua + str(atoms.count(ua))
        return mol_name

    def _load_basis(self):
        """Get the basis information needed to compute the AO values."""

        h5 = h5py.File(self.hdf5file, 'r')
        basis_grp = h5['molecule']['basis']
        self.basis = SimpleNamespace()

        self.basis.radial_type = basis_grp['radial_type'][()]
        self.basis.harmonics_type = basis_grp['harmonics_type'][()]

        self.basis.nao = int(basis_grp['nao'][()])
        self.basis.nmo = int(basis_grp['nmo'][()])

        self.basis.nshells = basis_grp['nshells'][()]
        self.basis.index_ctr = basis_grp['index_ctr'][()]

        self.basis.bas_exp = basis_grp['bas_exp'][()]
        self.basis.bas_coeffs = basis_grp['bas_coeffs'][()]

        self.basis.atom_coords_internal = basis_grp['atom_coords_internal'][(
        )]

        self.basis.TotalEnergy = basis_grp['TotalEnergy'][()]
        self.basis.mos = basis_grp['mos'][()]

        if self.basis.harmonics_type == 'cart':
            self.basis.bas_kr = basis_grp['bas_kr'][()]
            self.basis.bas_kx = basis_grp['bas_kx'][()]
            self.basis.bas_ky = basis_grp['bas_ky'][()]
            self.basis.bas_kz = basis_grp['bas_kz'][()]

        elif self.basis.harmonics_type == 'sph':
            self.basis.bas_n = basis_grp['bas_n'][()]
            self.basis.bas_l = basis_grp['bas_l'][()]
            self.basis.bas_m = basis_grp['bas_m'][()]

        h5.close()
        return self.basis

    # def load_mo_coeffs(self):
    #     """Get the molecule orbital coefficients."""

    #     h5 = h5py.File(self.hdf5file, 'r')
    #     return h5['molecule']['basis']['mos'][()]

    def print_total_energy(self):
        """Print the SCF energy of the molecule.

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis='sz')
            >>> mol.print_total_energy()
        """
        h5 = h5py.File(self.hdf5file, 'r')
        e = h5['molecule']['basis']['TotalEnergy'][()]
        print('== SCF Energy : ', e)
        h5.close()

    def _check_basis(self):
        """Check if the basis contains all the necessary fields."""

        names = ['bas_coeffs', 'bas_exp', 'nshells',
                 'atom_coords_internal', 'nao', 'nmo',
                 'index_ctr', 'mos', 'TotalEnergy']

        if self.basis.harmonics_type == 'cart':
            names += ['bas_kx', 'bas_ky', 'bas_kz', 'bas_kr']

        elif self.basis.harmonics_type == 'sph':
            names += ['bas_n', 'bas_l', 'bas_m']

        for n in names:
            if not hasattr(self.basis, n):
                raise ValueError(n, ' not in the basis namespace')

    def _load_hdf5(self, filename):
        """Load a molecule from hdf5

        Args:
            filename (str): path to the file to be loaded
        """

        # load the data
        load_from_hdf5(self, filename, 'molecule')

        # cast some of the important data type
        # should be done by the hdf5_utils in the future
        self.atoms = self.atoms.astype('U')
        self.basis.nao = int(self.basis.nao)
        self.basis.nmo = int(self.basis.nmo)

        cast_fn = {'nelec': int,
                   'nup': int,
                   'ndown': int,
                   'atoms': lambda x: x.astype('U'),
                   'atomic_nelec': lambda x: [int(i) for i in x]}

        for name, fn in cast_fn.items():
            self.__setattr__(name, fn(self.__getattribute__(name)))

        cast_fn = {'nao': int,
                   'nmo': int}

        for name, fn in cast_fn.items():
            self.basis.__setattr__(
                name, fn(self.basis.__getattribute__(name)))



import torch

from torch.autograd import grad, Variable


class WaveFunction(torch.nn.Module):

    def __init__(self, nelec, ndim, kinetic='auto', cuda=False):

        super(WaveFunction, self).__init__()

        self.ndim = ndim
        self.nelec = nelec
        self.ndim_tot = self.nelec * self.ndim
        self.kinetic = kinetic
        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

    def forward(self, x):
        ''' Compute the value of the wave function.
        for a multiple conformation of the electrons

        Args:
            parameters : variational param of the wf
            pos: position of the electrons

        Returns: values of psi
        '''

        raise NotImplementedError()

    def electronic_potential(self, pos):
        '''Compute the potential of the wf points
        Args:
            pos: position of the electron

        Returns: values of Vee * psi
        '''
        raise NotImplementedError()

    def nuclear_potential(self, pos):
        '''Compute the potential of the wf points
        Args:
            pos: position of the electron

        Returns: values of Ven * psi
        '''
        raise NotImplementedError()

    def nuclear_repulsion(self):
        '''Compute the nuclear repulsion term

        Returns: values of Vnn * psi
        '''
        raise NotImplementedError()

    def kinetic_energy(self, pos):
        '''Main switch for the kinetic energy.'''

        if self.kinetic == 'auto':
            return self.kinetic_energy_autograd(pos)

        elif self.kinetic == 'fd':
            return self.kinetic_energy_finite_difference(pos)

        else:
            raise ValueError(
                'kinetic %s not recognized' %
                self.kinetic)

    def kinetic_energy_autograd(self, pos):
        '''Compute the second derivative of the network
        output w.r.t the value of the input.

        This is to compute the value of the kinetic operator.

        Args:
            pos: position of the electron
            out : preomputed values of the wf at pos

        Returns:
            values of nabla^2 * Psi
        '''

        out = self.forward(pos)

        # compute the jacobian
        z = torch.ones_like(out)
        jacob = grad(out, pos,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        # compute the diagonal element of the Hessian
        z = Variable(torch.ones(jacob.shape[0])).to(self.device)
        hess = torch.zeros(jacob.shape[0]).to(self.device)

        for idim in range(jacob.shape[1]):

            tmp = grad(jacob[:, idim], pos,
                       grad_outputs=z,
                       only_inputs=True,
                       # create_graph is REQUIRED and
                       # is causing memory issues
                       # for large systems
                       create_graph=True)[0]

            hess += tmp[:, idim]

        return -0.5 * hess.view(-1, 1) / out

    def kinetic_energy_finite_difference(self, pos, eps=1E-3):
        '''Compute the second derivative of the network
        output w.r.t the value of the input using finite difference.

        This is to compute the value of the kinetic operator.

        Args:
            pos: position of the electron
            out : preomputed values of the wf at pos
            eps : psilon for numerical derivative
        Returns:
            values of nabla^2 * Psi
        '''

        nwalk = pos.shape[0]
        ndim = pos.shape[1]
        out = torch.zeros(nwalk, 1)

        pos_tmp = pos.clone()
        out = -2 * self.forward(pos_tmp)

        for icol in range(ndim):

            feps = out.clone()

            pos_tmp = pos.clone()
            pos_tmp[:, icol] += eps
            feps += self.forward(pos_tmp)

            pos_tmp = pos.clone()
            pos_tmp[:, icol] -= eps
            feps += self.forward(pos_tmp)

            out += feps / (eps**2)

        return -0.5 * out.view(-1, 1) / out

    def local_energy(self, pos):
        ''' local energy of the sampling points.'''

        ke = self.kinetic_energy(pos)

        return ke  \
            + self.nuclear_potential(pos)  \
            + self.electronic_potential(pos) \
            + self.nuclear_repulsion()

    def energy(self, pos):
        '''Total energy for the sampling points.'''
        return torch.mean(self.local_energy(pos))

    def variance(self, pos):
        '''Variance of the energy at the sampling points.'''
        return torch.var(self.local_energy(pos))

    def sampling_error(self, eloc):
        '''Compute the statistical uncertainty.
        Assuming the samples are uncorrelated.'''
        Npts = eloc.shape[0]
        return torch.sqrt(eloc.var() / Npts)

    def _energy_variance(self, pos):
        '''Return energy and variance.'''
        el = self.local_energy(pos)
        return torch.mean(el), torch.var(el)

    def _energy_variance_error(self, pos):
        '''Return energy variance and sampling error.'''
        el = self.local_energy(pos)
        return torch.mean(el), torch.var(el), self.sampling_error(el)

    def pdf(self, pos):
        '''density of the wave function.'''
        return (self.forward(pos)**2).reshape(-1)



import torch
from torch.autograd import Variable
from qmctorch.wavefunction import Orbital, Molecule
from pyscf import gto

import numpy as np
import unittest

import matplotlib.pyplot as plt

import os


class TestAOvalues(unittest.TestCase):

    def setUp(self):

        # define the molecule
        at = 'H 0 0 0; H 0 0 1'
        self.mol = Molecule(atom=at,
                            calculator='pyscf',
                            basis='sto-3g',
                            unit='bohr')

        self.m = gto.M(atom=at, basis='sto-3g', unit='bohr')

        # define the wave function
        self.wf = Orbital(self.mol)

        self.pos = torch.zeros(100, self.mol.nelec * 3)
        self.pos[:, 2] = torch.linspace(-5, 5, 100)

        self.pos = Variable(self.pos)
        self.pos.requires_grad = True
        self.iorb = 0
        self.x = self.pos[:, 2].detach().numpy()

    def test_ao(self):

        aovals = self.wf.ao(self.pos).detach().numpy()
        aovals_ref = self.m.eval_gto(
            'GTOval_cart', self.pos.detach().numpy()[:, :3])

        assert np.allclose(
            aovals[:, 0, self.iorb], aovals_ref[:, self.iorb])

    def test_ao_deriv(self):

        ip_aovals = self.wf.ao(
            self.pos, derivative=1).detach().numpy()
        ip_aovals_ref = self.m.eval_gto(
            'GTOval_ip_cart', self.pos.detach().numpy()[:, :3])
        ip_aovals_ref = ip_aovals_ref.sum(0)

        assert np.allclose(ip_aovals[:, 0, self.iorb],
                           ip_aovals_ref[:, self.iorb])

    def test_ao_hess(self):

        i2p_aovals = self.wf.ao(
            self.pos, derivative=2).detach().numpy()

        path = os.path.dirname(os.path.realpath(__file__))
        i2p_aovals_ref = np.loadtxt(path + '/hess_ao_h2.dat')

        assert np.allclose(
            i2p_aovals[:, 0, self.iorb], i2p_aovals_ref)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import pickle


def plot_energy(local_energy, e0=None, show_variance=False):
    """Plot the evolution of the energy

    Args:
        local_energy (np.ndarray): local energies along the trajectory
        e0 (float, optional): Target value for the energy. Defaults to None.
        show_variance (bool, optional): show the variance if True. Defaults to False.
    """

    fig = plt.figure()
    ax = fig.add_subplot(111)

    n = len(local_energy)
    epoch = np.arange(n)

    # get the variance
    energy = np.array([np.mean(e) for e in local_energy])
    variance = np.array([np.var(e) for e in local_energy])

    # get the mean value
    print("Energy   : %f " % np.mean(energy))
    print("Variance : %f " % np.std(energy))

    # plot
    ax.fill_between(epoch, energy - variance, energy +
                    variance, alpha=0.5, color='#4298f4')
    ax.plot(epoch, energy, color='#144477')
    if e0 is not None:
        ax.axhline(e0, color='black', linestyle='--')

    ax.grid()
    ax.set_xlabel('Number of epoch')
    ax.set_ylabel('Energy', color='black')

    if show_variance:
        ax2 = ax.twinx()
        ax2.plot(epoch, variance, color='blue')
        ax2.set_ylabel('variance', color='blue')
        ax2.tick_params(axis='y', labelcolor='blue')
        fig.tight_layout()

    plt.show()


def plot_data(observable, obsname):
    """Plot the evolution a given data

    Args:
        obs_dict (SimpleNamespace): namespace of observable
        obsname (str): name (key) of the desired observable
    """

    fig = plt.figure()
    ax = fig.add_subplot(111)

    data = np.array(observable.__getattribute__(obsname)).flatten()
    epoch = np.arange(len(data))
    ax.plot(epoch, data, color='#144477')

    if obsname + '.grad' in observable.__dict__.keys():
        data = np.array(observable.__getattribute__(
            obsname + '.grad')).flatten()
        ax2 = ax.twinx()
        ax2.plot(epoch, data, color='blue')
        ax2.set_ylabel('gradient', color='blue')
        ax2.tick_params(axis='y', labelcolor='blue')

    plt.show()


def plot_walkers_traj(eloc, walkers='mean'):
    """Plot the trajectory of all the individual walkers

    Args:
        obs (SimpleNamespace): Namespace of the observables
        walkers (int, str, optional): all, mean or index of a given walker Defaults to 'all'

    Returns:
        float :  Decorelation time
    """

    #eloc = np.array(obs.local_energy).squeeze(-1)

    nstep, nwalkers = eloc.shape

    celoc = np.cumsum(eloc, axis=0).T
    celoc /= np.arange(1, nstep + 1)

    var_decor = np.sqrt(np.var(np.mean(celoc, axis=1)))
    var = np.sqrt(np.var(celoc, axis=1) / (nstep - 1))

    Tc = (var_decor / var)**2

    if walkers is not None:
        # plt.subplot(1, 2, 1)

        if walkers == 'all':
            plt.plot(eloc, 'o', alpha=1 / nwalkers, c='grey')
            cmap = cm.hot(np.linspace(0, 1, nwalkers))
            for i in range(nwalkers):
                plt.plot(celoc.T[:, i], color=cmap[i])

        elif walkers == 'mean':
            plt.plot(eloc, 'o', alpha=1 / nwalkers, c='grey')
            plt.plot(np.mean(celoc.T, axis=1), linewidth=5)

        else:
            plt.plot(eloc[walkers, :], 'o',
                     alpha=1 / nwalkers, c='grey')
            plt.plot(celoc.T[traj_index, :])
        plt.grid()
        plt.xlabel('Monte Carlo Steps')
        plt.ylabel('Energy (Hartree)')
        # plt.subplot(1, 2, 2)
        # plt.hist(Tc)

    plt.show()

    return Tc


def plot_block(eloc):
    """Plot the blocking thingy

    Args:
        eloc (np.array): values of the local energy
    """

    nstep, nwalkers = eloc.shape
    max_block_size = nstep // 2

    evar = []
    for size in range(1, max_block_size):
        nblock = nstep // size
        tmp = np.copy(eloc[:size * nblock, :])
        tmp = tmp.reshape(nblock, size, nwalkers).mean(axis=1)
        evar.append(np.sqrt(np.var(tmp, axis=0) / (nblock - 1)))

    plt.plot(np.array(evar))
    plt.show()



__all__ = ['plot_energy', 'plot_data', 'plot_block',
           'plot_walkers_traj',
           'set_torch_double_precision',
           'set_torch_single_precision',
           'DataSet', 'Loss', 'OrthoReg',
           'dump_to_hdf5', 'load_from_hdf5',
           'register_extra_attributes']

from .plot_data import (plot_energy, plot_data, plot_block,
                        plot_walkers_traj)

from .torch_utils import (set_torch_double_precision,
                          set_torch_single_precision,
                          DataSet, Loss, OrthoReg)

from .hdf5_utils import (dump_to_hdf5, load_from_hdf5,
                         add_group_attr, register_extra_attributes)



import torch
from torch import nn


class ElectronDistance(nn.Module):

    def __init__(self, nelec, ndim):
        """Computes the electron-electron distances

        Args:
            nelec (int): number of electrons
            ndim (int): number of spatial dimensions 

        Examples::
            >>> edist = ElectronDistance(2,3)
            >>> pos = torch.tensor(500,6)
            >>> r = edist(pos)
            >>> dr = edist(pos,derivative=1)

        """

        super(ElectronDistance, self).__init__()
        self.nelec = nelec
        self.ndim = ndim

        _type_ = torch.get_default_dtype()
        if _type_ == torch.float32:
            self.eps = 1E-6
        elif _type_ == torch.float64:
            self.eps = 1E-16

    def forward(self, input, derivative=0):
        """Compute the pairwise distance between the electrons
        or its derivative. \n

        When required, the derivative is computed wrt to the first electron i.e.

        .. math::
            \\frac{dr_{ij}}{dx_i} 

        which is different from :    

        .. math::
            \\frac{d r_{ij}}{dx_j}

        Args:
            input (torch.tesnor): position of the electron \n
                                  size : Nbatch x [Nelec x Ndim]
            derivative (int, optional): degre of the derivative. \n
                                        Defaults to 0.

        Returns:
            torch.tensor: distance (or derivative) matrix \n
                          Nbatch x Nelec x Nelec if derivative = 0 \n
                          Nbatch x Ndim x  Nelec x Nelec if derivative = 1,2

        """

        # get the distance matrices
        input_ = input.view(-1, self.nelec, self.ndim)
        dist = self._get_distance_quadratic(input_)

        # eosilon on the diag needed for back prop
        eps_ = self.eps * \
            torch.diag(dist.new_ones(dist.shape[-1])).expand_as(dist)

        # extact the diagonal as diag can be negative someties
        # due to numerical noise
        diag = torch.diag_embed(
            torch.diagonal(
                dist, dim1=-1, dim2=-2))

        # remove diagonal and add eps for backprop
        dist = torch.sqrt(dist - diag + eps_)

        if derivative == 0:
            return dist

        elif derivative == 1:

            eps_ = self.eps * \
                torch.diag(dist.new_ones(
                    dist.shape[-1])).expand_as(dist)

            invr = (1. / (dist + eps_)).unsqueeze(1)
            diff_axis = input_.transpose(1, 2).unsqueeze(3)
            diff_axis = diff_axis - diff_axis.transpose(2, 3)
            return diff_axis * invr

        elif derivative == 2:

            eps_ = self.eps * \
                torch.diag(dist.new_ones(
                    dist.shape[-1])).expand_as(dist)
            invr3 = (1. / (dist**3 + eps_)).unsqueeze(1)
            diff_axis = input_.transpose(1, 2).unsqueeze(3)
            diff_axis = (diff_axis - diff_axis.transpose(2, 3))**2

            diff_axis = diff_axis[:, [
                [1, 2], [2, 0], [0, 1]], ...].sum(2)
            return (diff_axis * invr3)

    @staticmethod
    def _get_distance_quadratic(input):
        """Compute the distance following a quadratic expansion

        Arguments:
            input {torch.tensor} -- electron position [nbatch x nelec x ndim]

        Returns:
            torch.tensor -- distance matrices nbatch x nelec x ndim]
        """

        norm = (input**2).sum(-1).unsqueeze(-1)
        dist = (norm + norm.transpose(1, 2) - 2.0 *
                torch.bmm(input, input.transpose(1, 2)))
        return dist

    def _get_distance_diff(self, input):
        """Compute the distance following a difference scheme

        Arguments:
            input {torch.tensor} -- electron position [nbatch x nelec x ndim]

        Returns:
            torch.tensor -- distance matrices nbatch x nelec x ndim]
        """
        nbatch = input.shape[0]
        in1 = input.unsqueeze(1).expand(
            nbatch, self.nelec, self.nelec, self.ndim)
        in2 = input.unsqueeze(2).expand(
            nbatch, self.nelec, self.nelec, self.ndim)
        dist = torch.pow(in1 - in2, 2).sum(3)
        return dist



import torch
from types import SimpleNamespace
from tqdm import tqdm
import numpy as np

from ..utils import dump_to_hdf5, add_group_attr


class SolverBase(object):

    def __init__(self, wf=None, sampler=None,
                 optimizer=None, scheduler=None,
                 output=None):
        """Base Class for QMC solver 

        Args:
            wf (qmctorch.WaveFunction, optional): wave function. Defaults to None.
            sampler (qmctorch.sampler, optional): Sampler. Defaults to None.
            optimizer (torch.optim, optional): optimizer. Defaults to None.
            scheduler (torch.optim, optional): scheduler. Defaults to None.
            output (str, optional): hdf5 filename. Defaults to None.
        """

        self.wf = wf
        self.sampler = sampler
        self.opt = optimizer
        self.scheduler = scheduler
        self.cuda = False
        self.device = torch.device('cpu')
        self.task = None

        # penalty to orthogonalize the MO
        # see torch_utils.py
        self.ortho_mo = False

        # default task optimize the wave function
        self.configure(task='wf_opt')

        # resampling
        self.configure_resampling()

        # distributed model
        self.save_model = 'model.pth'

        # handles GPU availability
        if self.wf.cuda:
            self.device = torch.device('cuda')
            self.sampler.cuda = True
            self.sampler.walkers.cuda = True
        else:
            self.device = torch.device('cpu')

        self.hdf5file = output
        if output is None:
            basename = self.wf.mol.hdf5file.split('.')[0]
            self.hdf5file = basename + '_QMCTorch.hdf5'
        dump_to_hdf5(self, self.hdf5file)

    def configure_resampling(self, mode='update', resample_every=1, nstep_update=25):
        """Configure the resampling

        Args:
            mode (str, optional): method to resample : 'full', 'update', 'never' 
                                  Defaults to 'update'.
            resample_every (int, optional): Number of optimization steps between resampling
                                 Defaults to 1.
            nstep_update (int, optional): Number of MC steps in update mode. 
                                          Defaults to 25.
        """

        self.resampling_options = SimpleNamespace()
        valid_mode = ['never', 'full', 'update']
        if mode not in valid_mode:
            raise ValueError(
                mode, 'not a valid update method : ', valid_mode)

        self.resampling_options.mode = mode
        self.resampling_options.resample_every = resample_every
        self.resampling_options.nstep_update = nstep_update

    def configure(self, task='wf_opt', freeze=None):
        """Configure the optimization.

        Args:
            task (str, optional): Optimization task: 'wf_opt', 'geo_opt'.
                                  Defaults to 'wf_opt'.
            freeze (list, optional): list pf layers to freeze.
                                     Defaults to None.
        """

        self.task = task

        if task == 'geo_opt':
            self.configure_geo_opt()

        elif task == 'wf_opt':
            self.configure_wf_opt()

            self.freeze_parameters(freeze)

    def configure_geo_opt(self):
        """Configure the solver for geometry optimization."""

        # opt atom coordinate
        self.wf.ao.atom_coords.requires_grad = True

        # no ao opt
        self.wf.ao.bas_coeffs.requires_grad = False
        self.wf.ao.bas_exp.requires_grad = False

        # no jastrow opt
        self.wf.jastrow.weight.requires_grad = False

        # no mo opt
        for param in self.wf.mo.parameters():
            param.requires_grad = False

        # no ci opt
        self.wf.fc.weight.requires_grad = False

    def configure_wf_opt(self):
        """Configure the solver for wf optimization."""

        # opt all wf parameters
        self.wf.ao.bas_exp.requires_grad = True
        self.wf.ao.bas_coeffs.requires_grad = True
        for param in self.wf.mo.parameters():
            param.requires_grad = True
        self.wf.fc.weight.requires_grad = True
        self.wf.jastrow.weight.requires_grad = True

        # no opt the atom positions
        self.wf.ao.atom_coords.requires_grad = False

    def freeze_parameters(self, freeze):
        """Freeze the optimization of specified params.

        Args:
            freeze (list): list of param to freeze
        """
        if freeze is not None:
            if not isinstance(freeze, list):
                freeze = [freeze]

            for name in freeze:
                if name.lower() == 'ci':
                    self.wf.fc.weight.requires_grad = False

                elif name.lower() == 'mo':
                    for param in self.wf.mo.parameters():
                        param.requires_grad = False

                elif name.lower() == 'ao':
                    self.wf.ao.bas_exp.requires_grad = False
                    self.wf.ao.bas_coeffs.requires_grad = False

                elif name.lower() == 'jastrow':
                    self.wf.jastrow.weight.requires_grad = False

                else:
                    opt_freeze = ['ci', 'mo', 'ao', 'jastrow']
                    raise ValueError(
                        'Valid arguments for freeze are :', opt_freeze)

    def track_observable(self, obs_name):
        """define the observalbe we want to track

        Args:
            obs_name (list): list of str defining the observalbe.
                             Each str must correspond to a WaveFuncion method
        """

        # reset the Namesapce
        self.observable = SimpleNamespace()

        if 'local_energy' not in obs_name:
            obs_name += ['local_energy']

        if self.task == 'geo_opt' and 'geometry' not in obs_name:
            obs_name += ['geometry']

        for k in obs_name:
            self.observable.__setattr__(k, [])

        for key, p in zip(self.wf.state_dict().keys(),
                          self.wf.parameters()):
            if p.requires_grad:
                self.observable.__setattr__(key, [])
                self.observable.__setattr__(key+'.grad', [])

    def store_observable(self, pos, local_energy=None, ibatch=None, **kwargs):
        """store observale in the dictionary

        Args:
            obs_dict (dict): dictionary of the observalbe
            pos (torch.tensor): positions of th walkers
            local_energy (torch.tensor, optional): precomputed values of the local
                                           energy. Defaults to None
            ibatch (int): index of the current batch. Defaults to None
        """

        if self.wf.cuda and pos.device.type == 'cpu':
            pos = pos.to(self.device)

        for obs in self.observable.__dict__.keys():

            # store local energy
            if obs == 'local_energy' and local_energy is not None:
                data = local_energy.cpu().detach().numpy()

                if (ibatch is None) or (ibatch == 0):
                    self.observable.local_energy.append(data)
                else:
                    self.observable.local_energy[-1] = np.append(
                        self.observable.local_energy[-1], data)

            # store variational parameter
            elif obs in self.wf.state_dict():
                layer, param = obs.split('.')
                p = self.wf.__getattr__(layer).__getattr__(param)
                self.observable.__getattribute__(
                    obs).append(p.data.clone().numpy())

                if p.grad is not None:
                    self.observable.__getattribute__(obs +
                                                     '.grad').append(p.grad.clone().numpy())
                else:
                    self.observable.__getattribute__(obs +
                                                     '.grad').append(torch.zeros_like(p.data).numpy())

            # store any other defined method
            elif hasattr(self.wf, obs):
                func = self.wf.__getattribute__(obs)
                data = func(pos)
                if isinstance(data, torch.Tensor):
                    data = data.cpu().detach().numpy()
                self.observable.__getattribute__(obs).append(data)

    def print_observable(self, cumulative_loss, verbose=False):
        """Print the observalbe to csreen

        Args:
            cumulative_loss (float): current loss value
            verbose (bool, optional): print all the observables. Defaults to False
        """

        for k in self.observable.__dict__.keys():

            if k == 'local_energy':

                eloc = self.observable.local_energy[-1]
                e = np.mean(eloc)
                v = np.var(eloc)
                err = np.sqrt(v / len(eloc))
                print('energy   : %f +/- %f' % (e, err))
                print('variance : %f' % np.sqrt(v))

            elif verbose:
                print(
                    k + ' : ', self.observable.__getattribute__(k)[-1])
                print('loss %f' % (cumulative_loss))

    def resample(self, n, pos):
        """Resample the wave function

        Args:
            n (int): current epoch value
            nepoch (int): total number of epoch
            pos (torch.tensor): positions of the walkers

        Returns:
            (torch.tensor): new positions of the walkers
        """

        if self.resampling_options.mode != 'never':

            # resample the data
            if (n % self.resampling_options.resample_every == 0):

                # make a copy of the pos if we update
                if self.resampling_options.mode == 'update':
                    pos = pos.clone().detach().to(self.device)

                # start from scratch otherwise
                else:
                    pos = None

                # sample and update the dataset
                pos = self.sampler(self.wf.pdf, pos=pos)
                self.dataloader.dataset.data = pos

            # update the weight of the loss if needed
            if self.loss.use_weight:
                self.loss.weight['psi0'] = None

        return pos

    def single_point(self, hdf5_group='single_point'):
        """Performs a single point calculatin

        Args:
            hdf5_group (str, optional): hdf5 group where to store the data.
                                        Defaults to 'single_point'.

        Returns:
            SimpleNamespace: contains the local energy, positions, ...
        """

        # check if we have to compute and store the grads
        grad_mode = torch.no_grad()
        if self.wf.kinetic == 'auto':
            grad_mode = torch.enable_grad()

        with grad_mode:

            #  get the position and put to gpu if necessary
            pos = self.sampler(self.wf.pdf)
            if self.wf.cuda and pos.device.type == 'cpu':
                pos = pos.to(self.device)

            # compute energy/variance/error
            el = self.wf.local_energy(pos)
            e, s, err = torch.mean(el), torch.var(
                el), self.wf.sampling_error(el)

            # print data
            print('Energy   : ', e.detach().item(),
                  ' +/- ', err.detach().item())
            print('Variance : ', s.detach().item())

            # dump data to hdf5
            obs = SimpleNamespace(
                pos=pos,
                local_energy=el,
                energy=e,
                variance=s,
                error=err
            )
            dump_to_hdf5(obs,
                         self.hdf5file,
                         root_name=hdf5_group)
            add_group_attr(self.hdf5file, hdf5_group,
                           {'type': 'single_point'})

        return obs

    def save_checkpoint(self, epoch, loss, filename):
        torch.save({
            'epoch': epoch,
            'model_state_dict': self.wf.state_dict(),
            'optimzier_state_dict': self.opt.state_dict(),
            'loss': loss
        }, filename)
        return loss

    def _append_observable(self, key, data):
        """Append a new data point to observable key.

        Arguments:
            key {str} -- name of the observable
            data {} -- data
        """

        if key not in self.obs_dict.keys():
            self.obs_dict[key] = []
        self.obs_dict[key].append(data)

    def sampling_traj(self, pos, hdf5_group='sampling_trajectory'):
        """Compute the local energy along a sampling trajectory

        Args:
            pos (torch.tensor): positions of the walkers along the trajectory
            hdf5_group (str, optional): name of the group where to store the data.
                                        Defaults to 'sampling_trajecory'
        Returns:
            SimpleNamespace : contains energy/positions/
        """
        ndim = pos.shape[-1]
        p = pos.view(-1, self.sampler.nwalkers, ndim)
        el = []
        for ip in tqdm(p):
            el.append(self.wf.local_energy(ip).detach().numpy())

        el = np.array(el).squeeze(-1)
        obs = SimpleNamespace(local_energy=np.array(el), pos=pos)
        dump_to_hdf5(obs,
                     self.hdf5file, hdf5_group)

        add_group_attr(self.hdf5file, hdf5_group,
                       {'type': 'sampling_traj'})
        return obs

    def print_parameters(self, grad=False):
        for p in self.wf.parameters():
            if p.requires_grad:
                if grad:
                    print(p.grad)
                else:
                    print(p)

    def optimization_step(self, lpos):
        """Performs one optimization step

        Arguments:
            lpos {torch.tensor} -- positions of the walkers
        """

        if self.opt.lpos_needed:
            self.opt.step(lpos)
        else:
            self.opt.step()

    def save_traj(self, fname):
        """Save trajectory of geo_opt

        Args:
            fname (str): file name
        """
        f = open(fname, 'w')
        xyz = self.obs_dict['geometry']
        natom = len(xyz[0])
        nm2bohr = 1.88973
        for snap in xyz:
            f.write('%d \n\n' % natom)
            for at in snap:
                f.write('%s % 7.5f % 7.5f %7.5f\n' % (at[0],
                                                      at[1][0] /
                                                      nm2bohr,
                                                      at[1][1] /
                                                      nm2bohr,
                                                      at[1][2] / nm2bohr))
            f.write('\n')
        f.close()

    def run(self, nepoch, batchsize=None, loss='variance'):
        raise NotImplementedError()




import torch
from torch import nn
from .orbital_projector import OrbitalProjector


class SlaterPooling(nn.Module):

    """Applies a slater determinant pooling in the active space."""

    def __init__(self, configs, mol, cuda=False):
        """Computes the Sater determinants

        Args:
            configs (tuple): configuratin of the electrons
            mol (Molecule): Molecule instance
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.

        """
        super(SlaterPooling, self).__init__()

        self.configs = configs
        self.nconfs = len(configs[0])

        self.nmo = mol.basis.nmo
        self.nup = mol.nup
        self.ndown = mol.ndown

        self.orb_proj = OrbitalProjector(configs, mol)

        if cuda:
            self.device = torch.device('cuda')
            self.orb_proj.Pup = self.orb_proj.Pup.to(self.device)
            self.orb_proj.Pdown = self.orb_proj.Pdown.to(self.device)

    def forward(self, input, return_matrix=False):
        """Computes the values of the determinats

        Args:
            input (torch.tensor): MO matrices nbatc x nelec x nmo
            return_matrix (bool, optional): if true return the slater matrices 
                                            instead of their determinats. 
                                            Defaults to False.

        Returns:
            torch.tensor: slater determinants
        """

        mo_up, mo_down = self.orb_proj.split_orbitals(input)
        if return_matrix:
            return mo_up, mo_down
        else:
            return (
                torch.det(mo_up) *
                torch.det(mo_down)).transpose(0, 1)



import torch
from torch.autograd import Variable, grad
from pyscf import gto
from qmctorch.wavefunction import Molecule, Orbital
import unittest


def btrace(M):
    return torch.diagonal(M, dim1=-2, dim2=-1).sum(-1)


class OrbitalTest(Orbital):
    def __init__(self, mol):
        super(OrbitalTest, self).__init__(mol, use_jastrow=True)

    def first_der_autograd(self, x):
        """Compute the first derivative of the AO using autograd

        Args:
            x (torch.tensor): position of the electrons

        Returns:
            torch.tensor: jacbian of the AO
        """
        out = self.ao(x)

        z = Variable(torch.ones(out.shape))
        jacob = grad(out, x,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        return jacob

    def ao_trunk(self, x, n=4):
        """Extract the first 4x4 block of the AO

        Args:
            x (torch.tensor): positions
            n (int, optional): size of the block. Defaults to 4.

        Returns:
            torch.tensor: nxn block of the AO matrix
        """
        out = self.ao(x)
        return out[:, :n, :n]

    def second_der_autograd(self, pos, out=None):
        """Compute the second derivative of the AO using autograd

        Args:
            x (torch.tensor): position of the electrons

        Returns:
            torch.tensor: Hessian of the AO
        """

        out = self.ao(pos)

        # compute the jacobian
        z = Variable(torch.ones(out.shape))
        jacob = grad(out, pos,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        # compute the diagonal element of the Hessian
        z = Variable(torch.ones(jacob.shape[0]))
        hess = torch.zeros(jacob.shape)

        for idim in range(jacob.shape[1]):
            tmp = grad(jacob[:, idim], pos,
                       grad_outputs=z,
                       retain_graph=True,
                       only_inputs=True,
                       allow_unused=True)[0]

            hess[:, idim] = tmp[:, idim]

        return hess

    def second_der_autograd_mo(self, pos):
        """Compute the second derivative of the MOAO using autograd

        Args:
            x (torch.tensor): position of the electrons

        Returns:
            torch.tensor: Hessian of the AO
        """

        out = self.mo(self.ao(pos))

        # compute the jacobian
        z = Variable(torch.ones(out.shape))
        jacob = grad(out, pos,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        # compute the diagonal element of the Hessian
        z = Variable(torch.ones(jacob.shape[0]))
        hess = torch.zeros(jacob.shape)

        for idim in range(jacob.shape[1]):
            tmp = grad(jacob[:, idim], pos,
                       grad_outputs=z,
                       retain_graph=True,
                       only_inputs=True,
                       allow_unused=True)[0]

            hess[:, idim] = tmp[:, idim]

        return hess

    def first_der_trace(self, x, dAO=None):
        """Compute : Trace (AO^{-1} \nabla AO)

        Args:
            x (torch.tensor): position of the electrons
            dAO (torch.tensor, optional): precomputed values of the
                                          AO derivative. Defaults to None.

        Returns:
            torch tensor: values of Trace (AO^{-1} \nabla AO)
        """
        AO = self.ao(x)
        if dAO is None:
            dAO = self.ao(x, derivative=1)
        else:
            invAO = torch.inverse(AO)
        return btrace(invAO@dAO)

    def test_grad_autograd(self, pos):
        """Compute the jacobian of the AO block using autograd

        Args:
            pos (torch.tensor): position of the electrons

        Returns:
            torch tensor: jacobian of the AO
        """

        out = torch.det(self.ao_trunk(pos)).view(-1, 1)

        # compute the jacobian
        z = Variable(torch.ones(out.shape))
        jacob = grad(out, pos,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        return jacob.sum(1).view(-1, 1)

    def test_hess_autograd(self, pos):
        """Compute the hessian of the AO block using autograd

        Args:
            pos (torch.tensor): position of the electrons

        Returns:
            torch tensor: hessian of the AO
        """
        out = torch.det(self.ao_trunk(pos)).view(-1, 1)

        # compute the jacobian
        z = Variable(torch.ones(out.shape))
        jacob = grad(out, pos,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        # compute the diagonal element of the Hessian
        z = Variable(torch.ones(jacob.shape[0]))
        hess = torch.zeros(jacob.shape[0])

        for idim in range(jacob.shape[1]):
            tmp = grad(jacob[:, idim], pos,
                       grad_outputs=z,
                       retain_graph=True,
                       only_inputs=True,
                       allow_unused=True)[0]

            hess += tmp[:, idim]

        return hess.view(-1, 1)

    def test_kin_autograd(self, pos):
        """Compute the kinetic energy using autograd

        Args:
            pos (torch.tensor): electron position

        Returns:
            torch.tensor: kinetic energy values
        """

        out = self.mo(self.ao(pos))
        out = out[:, :4, :4]
        out = torch.det(out[:, :2, :2]) * torch.det(out[:, 2:, :2])
        out = out.view(-1)

        # compute the jacobian
        z = Variable(torch.ones(out.shape))
        jacob = grad(out, pos,
                     grad_outputs=z,
                     only_inputs=True,
                     create_graph=True)[0]

        # compute the diagonal element of the Hessian
        z = Variable(torch.ones(jacob.shape[0]))
        hess = torch.zeros(jacob.shape[0])

        for idim in range(jacob.shape[1]):
            tmp = grad(jacob[:, idim], pos,
                       grad_outputs=z,
                       retain_graph=True,
                       only_inputs=True,
                       allow_unused=True)[0]

            hess += tmp[:, idim]

        return hess.view(-1, 1)


class TestTrace(unittest.TestCase):

    def setUp(self):

        atom_str = 'C 0 0 -0.69; O 0 0 0.69'
        self.m = gto.M(atom=atom_str, basis='sto-3g', unit='bohr')
        self.mol = Molecule(atom=atom_str, calculator='pyscf',
                            basis='sto-3g', unit='bohr')

        # define the wave function
        self.wf = OrbitalTest(self.mol)
        self.x = 2 * torch.rand(5, 3 * self.mol.nelec) - 1.
        self.x.requires_grad = True

    def test_ao_der(self):
        """Test the values of the AO derivative."""
        dAO = self.wf.ao(self.x, derivative=1).sum()
        dAO_auto = self.wf.first_der_autograd(self.x).sum()
        assert(torch.allclose(dAO, dAO_auto))

    def test_ao_2der(self):
        """Test the values of the AO 2nd derivative."""
        d2AO = self.wf.ao(self.x, derivative=2).sum()
        d2AO_auto = self.wf.second_der_autograd(self.x).sum()
        assert(torch.allclose(d2AO, d2AO_auto))

    def test_mo_2der(self):
        """Test the values of the MO 2nd derivative."""
        d2MO = self.wf.mo(self.wf.ao(self.x, derivative=2)).sum()
        d2MO_auto = self.wf.second_der_autograd_mo(self.x).sum()
        assert(torch.allclose(d2MO, d2MO_auto))

    def test_trace(self):
        """Test the values jacobian and hessian with autograd and
        trace trick."""
        AO = self.wf.ao_trunk(self.x)
        iAO = torch.inverse(AO)

        dAO = self.wf.ao(self.x, derivative=1)
        d2AO = self.wf.ao(self.x, derivative=2)

        jac_auto = self.wf.test_grad_autograd(self.x)
        jac_trace = btrace(iAO@dAO[:, :4, :4]) * torch.det(AO)
        assert(torch.allclose(jac_auto.sum(), jac_trace.sum()))

        hess_auto = self.wf.test_hess_autograd(self.x)
        hess_trace = btrace(iAO@d2AO[:, :4, :4]) * torch.det(AO)
        assert(torch.allclose(hess_auto.sum(), hess_trace.sum()))

    def test_kinetic(self):
        """Test the values kinetic energy computed via autograd and
        trace trick."""

        # gives Nan on trvis servers when used with Jastrow
        kin_auto = self.wf.kinetic_energy_autograd(self.x)
        wfv = self.wf(self.x)
        kin_auto /= wfv

        kin_trace = self.wf.kinetic_energy_jacobi(
            self.x, return_local_energy=True)

        delta = kin_auto / kin_trace
        print(delta)

        # assert torch.allclose(delta, torch.ones_like(
        #     delta), atol=1e-3, rtol=1E-3)


if __name__ == "__main__":
    unittest.main()



from qmctorch.wavefunction import Molecule, Orbital
from qmctorch.utils import set_torch_double_precision
import numpy as np
import torch
import unittest
import itertools


class TestWaveFunction(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        self.wf = Orbital(mol, kinetic='auto', configs='cas(2,2)')
        self.pos = torch.tensor(np.random.rand(10, 6))
        self.pos.requires_grad = True

    def test_forward(self):

        wfvals = self.wf(self.pos)
        ref = torch.tensor([[0.1514],
                            [0.1427],
                            [0.2215],
                            [0.1842],
                            [0.0934],
                            [0.0697],
                            [0.0791],
                            [0.0963],
                            [0.1722],
                            [0.2988]])
        assert torch.allclose(wfvals.data, ref, rtol=1E-4, atol=1E-4)

    def test_local_energy(self):

        eloc_auto = self.wf.local_energy(self.pos)
        eloc_jac = self.wf.local_energy_jacobi(self.pos)
        ref = torch.tensor([[-1.3205],
                            [-1.1108],
                            [-1.2487],
                            [-1.0767],
                            [-0.4658],
                            [-0.2524],
                            [-0.6561],
                            [-0.5626],
                            [-1.1561],
                            [-1.9689]])

        assert torch.allclose(
            eloc_auto.data, ref, rtol=1E-4, atol=1E-4)

        assert torch.allclose(
            eloc_auto.data, eloc_jac.data, rtol=1E-4, atol=1E-4)

    def test_kinetic_energy(self):

        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)
        ref = torch.tensor([[0.9461],
                            [0.4120],
                            [2.1962],
                            [1.3390],
                            [0.0657],
                            [0.2884],
                            [0.5593],
                            [0.1747],
                            [1.1705],
                            [3.3116]])

        assert torch.allclose(
            ejac.data, ref, rtol=1E-4, atol=1E-4)

        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)

        efd = self.wf.kinetic_energy_finite_difference(self.pos)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import os
import shutil
import warnings
from .calculator_base import CalculatorBase
import h5py
from types import SimpleNamespace

try:
    from scm import plams
except ModuleNotFoundError:
    warnings.warn('scm python module not found')


class CalculatorADF(CalculatorBase):

    def __init__(self, atoms, atom_coords, basis, scf, units, molname):

        CalculatorBase.__init__(
            self, atoms, atom_coords, basis, scf, units, molname, 'adf')

    def run(self):
        """Run the calculation using ADF."""

        # path needed for the calculation
        wd = ''.join(self.atoms) + '_' + self.basis_name
        t21_name = wd + '.t21'
        plams_wd = './plams_workdir'
        t21_path = os.path.join(
            plams_wd, os.path.join(wd, t21_name))

        # configure plams and run the calculation
        self.init_plams()
        mol = self.get_plams_molecule()
        sett = self.get_plams_settings()
        job = plams.ADFJob(molecule=mol, settings=sett, name=wd)
        job.run()

        # extract the data to hdf5
        basis = self.get_basis_data(t21_path)
        shutil.rmtree(plams_wd)

        return basis

    def init_plams(self):
        """Init PLAMS."""
        plams.init()
        plams.config.log.stdout = -1
        plams.config.erase_workdir = True

    def get_plams_molecule(self):
        """Returns a plams molecule object."""
        mol = plams.Molecule()
        for at, xyz in zip(self.atoms, self.atom_coords):
            mol.add_atom(plams.Atom(symbol=at, coords=tuple(xyz)))
        return mol

    def get_plams_settings(self):
        """Returns a plams setting object."""

        sett = plams.Settings()
        sett.input.basis.type = self.basis_name.upper()
        sett.input.basis.core = 'None'
        sett.input.symmetry = 'nosym'
        sett.input.XC.HartreeFock = ''

        # correct unit
        if self.units == 'angs':
            sett.input.units.length = 'Angstrom'
        elif self.units == 'bohr':
            sett.input.units.length = 'Bohr'

        # total energy
        sett.input.totalenergy = True

        return sett

    def get_basis_data(self, kffile):
        """Save the basis information needed to compute the AO values."""

        kf = plams.KFFile(kffile)

        basis = SimpleNamespace()

        basis.TotalEnergy = kf.read('Total Energy', 'Total energy')
        basis.radial_type = 'sto'
        basis.harmonics_type = 'cart'

        nao = kf.read('Basis', 'naos')
        nmo = kf.read('A', 'nmo_A')
        basis.nao = nao
        basis.nmo = nmo

        # number of bas per atom type
        nbptr = kf.read('Basis', 'nbptr')

        # number of atom per atom typ
        nqptr = kf.read('Geometry', 'nqptr')
        atom_type = kf.read('Geometry', 'atomtype').split()

        # number of bas per atom type
        nshells = np.array([nbptr[i] - nbptr[i - 1]
                            for i in range(1, len(nbptr))])

        # kx/ky/kz/kr exponent per atom type
        bas_kx = np.array(kf.read('Basis', 'kx'))
        bas_ky = np.array(kf.read('Basis', 'ky'))
        bas_kz = np.array(kf.read('Basis', 'kz'))
        bas_kr = np.array(kf.read('Basis', 'kr'))

        # bas exp/coeff/norm per atom type
        bas_exp = np.array(kf.read('Basis', 'alf'))
        bas_norm = np.array(kf.read('Basis', 'bnorm'))

        basis_nshells = []
        basis_bas_kx, basis_bas_ky, basis_bas_kz = [], [], []
        basis_bas_kr = []
        basis_bas_exp, basis_bas_norm = [], []

        for iat, at in enumerate(atom_type):

            number_copy = nqptr[iat + 1] - nqptr[iat]
            idx_bos = list(range(nbptr[iat] - 1, nbptr[iat + 1] - 1))

            basis_nshells += [nshells[iat]] * number_copy

            basis_bas_kx += list(bas_kx[idx_bos]) * number_copy
            basis_bas_ky += list(bas_ky[idx_bos]) * number_copy
            basis_bas_kz += list(bas_kz[idx_bos]) * number_copy
            basis_bas_kr += list(bas_kr[idx_bos]) * number_copy
            basis_bas_exp += list(bas_exp[idx_bos]) * number_copy
            basis_bas_norm += list(
                bas_norm[idx_bos]) * number_copy

        basis.nshells = basis_nshells
        basis.index_ctr = np.arange(nao)

        basis.bas_kx = np.array(basis_bas_kx)
        basis.bas_ky = np.array(basis_bas_ky)
        basis.bas_kz = np.array(basis_bas_kz)
        basis.bas_kr = np.array(basis_bas_kr)

        basis.bas_exp = np.array(basis_bas_exp)
        basis.bas_coeffs = np.ones_like(basis_bas_exp)
        basis.bas_norm = np.array(basis_bas_norm)

        basis.atom_coords_internal = np.array(
            kf.read('Geometry', 'xyz')).reshape(-1, 3)

        # Molecular orbitals
        mos = np.array(kf.read('A', 'Eigen-Bas_A'))
        mos = mos.reshape(nmo, nao).T
        mos = self.normalize_columns(mos)
        basis.mos = mos

        return basis



import torch
from torch.autograd import Variable, grad, gradcheck
from qmctorch.wavefunction.jastrow import TwoBodyJastrowFactor
import unittest

torch.set_default_tensor_type(torch.DoubleTensor)


def hess(out, pos):
    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


class TestJastrow(unittest.TestCase):

    def setUp(self):
        self.nup, self.ndown = 2, 2
        self.nelec = self.nup + self.ndown
        self.jastrow = TwoBodyJastrowFactor(self.nup, self.ndown)
        self.nbatch = 5

        self.pos = torch.rand(self.nbatch, self.nelec * 3)
        self.pos.requires_grad = True

    def test_grad_distance(self):

        r = self.jastrow.edist(self.pos)
        dr = self.jastrow.edist(self.pos, derivative=1)
        dr_grad = grad(
            r,
            self.pos,
            grad_outputs=torch.ones_like(r))[0]
        gradcheck(self.jastrow.edist, self.pos)

        assert(torch.allclose(dr.sum(), dr_grad.sum(), atol=1E-5))

    def test_grad_jastrow(self):

        val = self.jastrow(self.pos)
        dval = self.jastrow(self.pos, derivative=1)
        dval_grad = grad(
            val,
            self.pos,
            grad_outputs=torch.ones_like(val))[0]
        gradcheck(self.jastrow, self.pos)

        assert torch.allclose(dval, dval_grad.view(
            self.nbatch, self.nelec, 3).sum(2))
        assert(torch.allclose(dval.sum(), dval_grad.sum()))

    def test_hess_jastrow(self):

        val = self.jastrow(self.pos)
        d2val_grad = hess(val, self.pos)
        d2val = self.jastrow(self.pos, derivative=2)

        # still an issue there :(
        assert torch.allclose(d2val, d2val_grad.view(
            self.nbatch, self.nelec, 3).sum(2))
        assert(torch.allclose(d2val.sum(), d2val_grad.sum()))


if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn

from .orbital_projector import OrbitalProjector


def btrace(M):
    return torch.diagonal(M, dim1=-2, dim2=-1).sum(-1)


def bproj(M, P):
    return P.transpose(1, 2) @ M @ P


class KineticPooling(nn.Module):

    def __init__(self, configs, mol, cuda=False):
        """Computes the kinetic energy using the Jacobi formula

        Args:
            configs (list): list of slater determinants
            mol (Molecule): instance of a Molecule object
            cuda (bool, optional): GPU ON/OFF. Defaults to False.
        """

        super(KineticPooling, self).__init__()

        self.configs = configs
        self.nconfs = len(configs[0])

        self.nup = mol.nup
        self.ndown = mol.ndown
        self.nelec = self.nup + self.ndown

        self.orb_proj = OrbitalProjector(configs, mol)

        if cuda:
            self.device = torch.device('cuda')
            self.orb_proj.Pup = self.orb_proj.Pup.to(self.device)
            self.orb_proj.Pdown = self.orb_proj.Pdown.to(self.device)

    def forward(self, MO, d2MO, dJdMO=None, d2JMO=None):
        """Compute the kinetic energy using the trace trick for a product of spin up/down determinant.

        .. math::
            -\\frac{1}{2} \Delta \Psi = -\\frac{1}{2}  D_{up} D_{down}  
            ( \Delta_{up} D_{up}  + \Delta_{down} D_{down} )

        Args:
            MO (torch.tensor): matrix of MO vals(Nbatch, Nelec, Nmo)
            d2MO (torch.tensor): matrix of :math:`\Delta` MO vals(Nbatch, Nelec, Nmo)
            dJdMO (torch.tensor, optional): matrix of the :math:`\\frac{\\nabla J}{J} \\nabla MO`. Defaults to None.
            d2JMO (torch.tensor, optional): matrix of the :math:`\\frac{\Delta J}{J} MO`. Defaults to None.

        Returns:
            torch.tensor: kinetic energy
        """

        # shortcut up/down matrices
        Aup, Adown = self.orb_proj.split_orbitals(MO)
        if dJdMO is None and d2JMO is None:
            Bup, Bdown = self.orb_proj.split_orbitals(d2MO)
        else:
            Bup, Bdown = self.orb_proj.split_orbitals(
                d2MO + 2 * dJdMO + d2JMO)

        # inverse of MO matrices
        iAup = torch.inverse(Aup)
        iAdown = torch.inverse(Adown)

        # determinant product
        det_prod = torch.det(Aup) * torch.det(Adown)

        # kinetic terms
        kinetic = -0.5 * (btrace(iAup@Bup) +
                          btrace(iAdown@Bdown)) * det_prod

        # reshape
        kinetic = kinetic.transpose(0, 1)
        det_prod = det_prod.transpose(0, 1)

        return kinetic, det_prod



import torch
from torch import nn
from .electron_distance import ElectronDistance
from ..utils import register_extra_attributes


class TwoBodyJastrowFactor(nn.Module):

    def __init__(self, nup, ndown, w=1., cuda=False):
        """Computes the Pade-Jastrow factor

        .. math::
            J = \prod_{i<j} \exp(B_{ij}) \quad \quad \\text{with} \quad \quad
            B_{ij} = \\frac{w_0 r_{i,j}}{1 + w r_{i,j}}

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            w (float, optional): Value of the variational parameter. Defaults to 1..
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super(TwoBodyJastrowFactor, self).__init__()

        self.nup = nup
        self.ndown = ndown
        self.nelec = nup + ndown
        self.ndim = 3

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

        self.weight = nn.Parameter(torch.tensor([w]))
        self.weight.requires_grad = True

        bup = torch.cat((0.25 * torch.ones(nup, nup), 0.5 *
                         torch.ones(nup, ndown)), dim=1)

        bdown = torch.cat((0.5 * torch.ones(ndown, nup), 0.25 *
                           torch.ones(ndown, ndown)), dim=1)

        self.static_weight = torch.cat(
            (bup, bdown), dim=0).to(
            self.device)

        self.edist = ElectronDistance(self.nelec, self.ndim)

        register_extra_attributes(self, ['weight'])

    def _to_device(self):
        """Export the non parameter variable to the device."""

        self.device = torch.device('cuda')
        self.to(self.device)
        attrs = ['static_weight']
        for at in attrs:
            self.__dict__[at] = self.__dict__[at].to(self.device)

    def forward(self, pos, derivative=0, jacobian=True):
        """Compute the Jastrow factors.

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                            Defaults to 0.
            jacobian (bool, optional): Return the jacobian (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

        Returns:
            torch.tensor: value of the jastrow parameter for all confs
        """
        if not jacobian:
            assert(derivative == 1)

        size = pos.shape
        assert size[1] == self.nelec * self.ndim
        r = self.edist(pos)
        jast = self._get_jastrow_elements(r)

        if derivative == 0:

            return self._prod_unique_pairs(jast)

        elif derivative == 1:
            dr = self.edist(pos, derivative=1)
            return self._jastrow_derivative(r, dr, jast, jacobian)

        elif derivative == 2:
            dr = self.edist(pos, derivative=1)
            d2r = self.edist(pos, derivative=2)
            return self._jastrow_second_derivative(r, dr, d2r, jast)

    def _jastrow_derivative(self, r, dr, jast, jacobian):
        """Compute the value of the derivative of the Jastrow factor

        Args:
            r (torch.tensor): ee distance matrix Nbatch x Nelec x Nelec
            jast (torch.tensor): values of the ajstrow elements
                                 Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: gradient of the jastrow factors
                          Nbatch x Nelec x Ndim
        """
        if jacobian:
            djast = self._get_der_jastrow_elements(r, dr).sum(1)
            prod_val = self._prod_unique_pairs(jast)

        else:
            djast = self._get_der_jastrow_elements(r, dr)
            prod_val = self._prod_unique_pairs(jast).unsqueeze(-1)

        return (self._sum_unique_pairs(djast, axis=-1) -
                self._sum_unique_pairs(djast, axis=-2)) * prod_val

    def _jastrow_second_derivative(self, r, dr, d2r, jast):
        """Compute the value of the pure 2nd derivative of the Jastrow factor

        Args:
            r (torch.tensor): ee distance matrix Nbatch x Nelec x Nelec
            jast (torch.tensor): values of the ajstrow elements
                                 Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: diagonal hessian of the jastrow factors
                          Nbatch x Nelec x Ndim
        """

        # pure second derivative terms
        prod_val = self._prod_unique_pairs(jast)
        d2jast = self._get_second_der_jastrow_elements(
            r, dr, d2r).sum(1)
        hess_jast = 0.5 * (self._sum_unique_pairs(d2jast, axis=-1)
                           + self._sum_unique_pairs(d2jast, axis=-2))

        # mixed terms
        djast = (self._get_der_jastrow_elements(r, dr))  # .sum(1)
        hess_jast += self._partial_derivative(
            djast, out_mat=hess_jast)

        return hess_jast * prod_val

    def _get_jastrow_elements(self, r):
        """Get the elements of the jastrow matrix :
        .. math::
            out_{i,j} = \exp{ \frac{b r_{i,j}}{1+b'r_{i,j}} }

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the jastrow elements
                          Nbatch x Nelec x Nelec
        """
        return torch.exp(self._compute_kernel(r))

    def _compute_kernel(self, r):
        """ Get the jastrow kernel.
        .. math::
            B_{ij} = \frac{b r_{i,j}}{1+b'r_{i,j}}

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: matrix of the jastrow kernels
                          Nbatch x Nelec x Nelec
        """
        return self.static_weight * r / (1.0 + self.weight * r)

    def _get_der_jastrow_elements(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels
        wrt to the first electrons

        .. math::

            d B_{ij} / d k_i =  d B_{ij} / d k_j  = - d B_{ji} / d k_i

            out_{k,i,j} = A1 + A2
            A1_{kij} = w0 \frac{dr_{ij}}{dk_i} / (1 + w r_{ij})
            A2_{kij} = - w0 w' r_{ij} \frac{dr_{ij}}{dk_i} / (1 + w r_{ij})^2

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the derivative of the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        r_ = r.unsqueeze(1)
        denom = 1. / (1.0 + self.weight * r_)
        a = self.static_weight * dr * denom
        b = - self.static_weight * self.weight * r_ * dr * denom**2

        return (a + b)

    def _get_second_der_jastrow_elements(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels
        wrt to the first electron

        .. math ::

            d^2 B_{ij} / d k_i^2 =  d^2 B_{ij} / d k_j^2 = d^2 B_{ji} / d k_i^2

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec
            d2r (torch.tensor): matrix of the 2nd derivative of 
                                the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the pure 2nd derivative of 
                          the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        r_ = r.unsqueeze(1)
        denom = 1. / (1.0 + self.weight * r_)
        denom2 = denom**2
        dr_square = dr**2
        a = self.static_weight * d2r * denom
        b = -2 * self.static_weight * self.weight * dr_square * denom2
        c = - self.static_weight * self.weight * r_ * d2r * denom2
        d = 2 * self.static_weight * self.weight**2 * r_ * dr_square * denom**3

        e = self._get_der_jastrow_elements(r, dr)

        return a + b + c + d + e**2

    def _partial_derivative(self, djast, out_mat=None):
        """Get the product of the mixed second deriative terms.

        .. math ::

            d B_{ij} / d x_i * d B_{kl} / d x_k

        Args:
            djast (torch.tensor): first derivative of the jastrow kernels
            out_mat (torch.tensor, optional): output matrix. Defaults to None.

        Returns:
            torch.tensor:
        """

        if out_mat is None:
            nbatch = djast.shape[0]
            out_mat = torch.zeros(nbatch, self.nelec)

        for idx in range(self.nelec):

            index_pairs = [(idx, j, 1) for j in range(
                idx + 1, self.nelec)] + [(j, idx, -1) for j in range(0, idx)]

            for p1 in range(len(index_pairs) - 1):
                i1, j1, w1 = index_pairs[p1]
                for p2 in range(p1 + 1, len(index_pairs)):
                    i2, j2, w2 = index_pairs[p2]

                    d1 = djast[..., i1, j1] * w1
                    d2 = djast[..., i2, j2] * w2

                    out_mat[..., idx] += (d1 * d2).sum(1)

        return out_mat

    def _prod_unique_pairs(self, mat, not_el=None):
        """Compute the product of the lower mat elements

        Args:
            mat (torch.tensor): input matrix [..., N x N]
            not_el (tuple(i,j), optional):
                single element(s) to exclude of the product.
                Defaults to None.

        Returns:
            torch.tensor : value of the product
        """

        mat_cpy = mat.clone()
        if not_el is not None:

            if not isinstance(not_el, list):
                not_el = [not_el]

            for _el in not_el:
                i, j = _el
                mat_cpy[..., i, j] = 1

        return mat_cpy[..., torch.tril(torch.ones(
            self.nelec, self.nelec)) == 0].prod(1).view(-1, 1)

    def _sum_unique_pairs(self, mat, axis=None):
        """Sum the unique pairs of the lower triangluar matrix

        Args:
            mat (torch.tensor): input matrix [..., N x N]
            axis (int, optional): index of the axis to sum. Defaults to None.

        Returns:
            torch.tensor:
        """

        mat_cpy = mat.clone()
        mat_cpy[..., torch.tril(torch.ones(
            self.nelec, self.nelec)) == 1] = 0

        if axis is None:
            return mat_cpy.sum()
        else:
            return mat_cpy.sum(axis)



import torch


class OrbitalProjector(object):

    def __init__(self, configs, mol):
        """Project the MO matrix in Slater Matrices

        Args:
            configs (list): configurations of the slater determinants
            mol (Molecule): Molecule object
        """

        self.configs = configs
        self.nconfs = len(configs[0])
        self.nmo = mol.basis.nmo
        self.nup = mol.nup
        self.ndown = mol.ndown

        self.Pup, self.Pdown = self.get_projectors()

    def get_projectors(self):
        """Get the projectors of the conf in the CI expansion

        Returns:
            torch.tensor, torch.tensor : projectors
        """

        Pup = torch.zeros(self.nconfs, self.nmo, self.nup)
        Pdown = torch.zeros(self.nconfs, self.nmo, self.ndown)

        for ic, (cup, cdown) in enumerate(
                zip(self.configs[0], self.configs[1])):

            for _id, imo in enumerate(cup):
                Pup[ic][imo, _id] = 1.

            for _id, imo in enumerate(cdown):
                Pdown[ic][imo, _id] = 1.

        return Pup.unsqueeze(1), Pdown.unsqueeze(1)

    def split_orbitals(self, mo):
        """Split the orbital  matrix in multiple slater matrices

        Args:
            mo (torch.tensor): molecular orbital matrix

        Returns:
            torch.tensor: all slater matrices
        """
        return mo[:, :self.nup, :] @ self.Pup, mo[:,
                                                  self.nup:, :] @ self.Pdown
